/* 
 * ccolor.c: color conversion and downsampling routines
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */

#define J3D_INTERNALS
#include <stdlib.h>
#include "j3dlib.h"

/* Private subobject */
typedef struct {
    struct j3d_c_color_converter pub; /* public fields */

    int * rgb_ycc_tab;		/* => table for RGB to YCbCr conversion */
} my_color_converter;
typedef my_color_converter * my_cconvert_ptr;




/**************** RGB -> YCbCr conversion: most common case **************/

/*
 * YCbCr is defined per CCIR 601-1, except that Cb and Cr are
 * normalized to the range 0..MAXJSAMPLE rather than -0.5 .. 0.5.
 * The conversion equations to be implemented are therefore
 *	Y  =  0.29900 * R + 0.58700 * G + 0.11400 * B
 *	Cb = -0.16874 * R - 0.33126 * G + 0.50000 * B  + CENTERJSAMPLE
 *	Cr =  0.50000 * R - 0.41869 * G - 0.08131 * B  + CENTERJSAMPLE
 * (These numbers are derived from TIFF 6.0 section 21, dated 3-June-92.)
 * Note: older versions of the IJG code used a zero offset of MAXJSAMPLE/2,
 * rather than CENTERJSAMPLE, for Cb and Cr.  This gave equal positive and
 * negative swings for Cb/Cr, but meant that grayscale values (Cb=Cr=0)
 * were not represented exactly.  Now we sacrifice exact representation of
 * maximum red and maximum blue in order to get exact grayscales.
 *
 * To avoid floating-point arithmetic, we represent the fractional constants
 * as integers scaled up by 2^16 (about 4 digits precision); we have to divide
 * the products by 2^16, with appropriate rounding, to get the correct answer.
 *
 * For even more speed, we avoid doing any multiplications in the inner loop
 * by precalculating the constants times R,G,B for all possible values.
 * For 8-bit JSAMPLEs this is very reasonable (only 256 entries per table);
 * for 12-bit samples it is still acceptable.  It's not very reasonable for
 * 16-bit samples, but if you want lossless storage you shouldn't be changing
 * colorspace anyway.
 * The CENTERJSAMPLE offsets and the rounding fudge-factor of 0.5 are included
 * in the tables to save adding them separately in the inner loop.
 */

#define MAXJSAMPLE      255
#define CENTERJSAMPLE	128

#define SCALEBITS	16	/* speediest right-shift on some machines */
#define CBCR_OFFSET	((int) CENTERJSAMPLE << SCALEBITS)
#define ONE_HALF	((int) 1 << (SCALEBITS-1))
#define FIX(x)		((int) ((x) * (1L<<SCALEBITS) + 0.5))

/* We allocate one big table and divide it up into eight parts, instead of
 * doing eight alloc_small requests.  This lets us use a single table base
 * address, which can be held in a register in the inner loops on many
 * machines (more than can hold all eight addresses, anyway).
 */

#define R_Y_OFF		0			/* offset to R => Y section */
#define G_Y_OFF		(1*(MAXJSAMPLE+1))	/* offset to G => Y section */
#define B_Y_OFF		(2*(MAXJSAMPLE+1))	/* etc. */
#define R_CB_OFF	(3*(MAXJSAMPLE+1))
#define G_CB_OFF	(4*(MAXJSAMPLE+1))
#define B_CB_OFF	(5*(MAXJSAMPLE+1))
#define R_CR_OFF	B_CB_OFF		/* B=>Cb, R=>Cr are the same */
#define G_CR_OFF	(6*(MAXJSAMPLE+1))
#define B_CR_OFF	(7*(MAXJSAMPLE+1))
#define TABLE_SIZE	(8*(MAXJSAMPLE+1))




/*
 * color downsampling for all downsampling ratios
 *
 * ToDo: optimization. (low priority, it will not be used very often)
 *
 */
static void downsample(j3d_color_component buf, unsigned int width, unsigned int height, unsigned int factor) {
    register unsigned int inptr, outptr;
    register unsigned int val;
    register unsigned int x, y;
    register unsigned int factor_sqr = factor*factor;

    /* assumes image can be divided by factor !*/
    width /= factor;
    height /= factor;

    for(outptr = 0; outptr < height * width; outptr++) {
        inptr = outptr * factor;

        /* process block of size factor*factor */
        val = 0; 
        for (y = 0; y < factor; y++) {
            for (x = 0; x < factor; x++) {
                val += buf[inptr + x];
            }
            inptr += width;
        }
        buf[outptr] = val / factor_sqr;
    }  
}



/*
 * color downsampling for the common case of 2:1 ratio (factor=2)
 *
 * writes output into the input buffer
 *
 * A note about the "bias" calculations: when rounding fractional values to
 * integer, we do not want to always round 0.5 up to the next integer.
 * If we did that, we'd introduce a noticeable bias towards larger values.
 * Instead, this code is arranged so that 0.5 will be rounded up or down at
 * alternate pixel locations (a simple ordered dither pattern).
 *
 */
static void downsample_2(j3d_color_component buf, unsigned int width, unsigned int height) {
    unsigned int x, y;

    register unsigned int inptr0, inptr1, outptr;
    register int bias;
    register unsigned int newwidth, newheight;

    /* assumes image can be divided by 2 !*/
    newwidth = width >> 1;
    newheight = height >> 1;

    inptr0 = 0;
    inptr1 = width;
    outptr = 0;

    for(y = 0; y < newheight; y++) {
        bias = 1;
        for(x = 0; x < newwidth; x++) {
            /* calculate mean value of 2*2 block */
            buf[outptr++] = (j3d_sample) ((buf[inptr0] + buf[inptr0 + 1] +
                                           buf[inptr1] + buf[inptr1 + 1] +
                                           bias) >> 2);
            bias ^= 3; /* 1=>2, 2=>1 */
            inptr0 += 2;
            inptr1 += 2;
        }
        inptr0 += width;
        inptr1 += width;
    }  
}



/*
 * color conversion and downsampling.
 * bgr scheme is used in bmp images.
 *
 * ToDo: efficient algorithm which does both in one pass
 */
static void bgr_ycc_convert(j3d_compress_ptr cinfo, j3d_rgb_image input_buf, j3d_image output_buf) {
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  
    int r, g, b;
    int * ctab = cconvert->rgb_ycc_tab;
    unsigned int incount, outcount;
    j3d_color_component comp0, comp1, comp2;
    unsigned int imagesize = (cinfo->width * cinfo->height);
    
    
    /* color conversion */
    incount = 0;
    comp0 = output_buf[0];
    comp1 = output_buf[1];
    comp2 = output_buf[2];

    
    for(outcount = 0; outcount < imagesize; outcount++) {
        /* this is the bmp order */
        b = input_buf[incount++]; 
        g = input_buf[incount++]; 
        r = input_buf[incount++]; 
        

        /* Y */
        comp0[outcount] = (j3d_sample)
            ((ctab[r+R_Y_OFF] + ctab[g+G_Y_OFF] + ctab[b+B_Y_OFF])
             >> SCALEBITS);

        /* Cb */
        comp1[outcount] = (j3d_sample)
            ((ctab[r+R_CB_OFF] + ctab[g+G_CB_OFF] + ctab[b+B_CB_OFF])
             >> SCALEBITS);

        /* Cr */
        comp2[outcount] = (j3d_sample)
            ((ctab[r+R_CR_OFF] + ctab[g+G_CR_OFF] + ctab[b+B_CR_OFF])
             >> SCALEBITS);
    }


    /* color downsampling */

    /* Y */
    if(cinfo->downsampling_factor[0] > 1) {
        downsample(comp0, cinfo->width, cinfo->height, cinfo->downsampling_factor[0]); 
    }

    /* Cb */
    if(cinfo->downsampling_factor[1] > 1) {
        if(cinfo->downsampling_factor[1] == 2) {
            downsample_2(comp1, cinfo->width, cinfo->height);
        } else {
            downsample(comp1, cinfo->width, cinfo->height, cinfo->downsampling_factor[1]);
        }
    }
        
    /* Cr */
    if(cinfo->downsampling_factor[2] > 1) {
        if(cinfo->downsampling_factor[2] == 2) {
            downsample_2(comp2, cinfo->width, cinfo->height);
        } else {
            downsample(comp2, cinfo->width, cinfo->height, cinfo->downsampling_factor[2]);
        }
    }
}



void j3d_init_c_color_converter(j3d_compress_ptr cinfo) {
    my_cconvert_ptr cconvert;
    int i;
    int * rgb_ycc_tab;
    cconvert = (my_cconvert_ptr) malloc(sizeof(my_color_converter));

    
    /* Allocate and fill in the conversion tables. */
    rgb_ycc_tab = (int*) calloc(TABLE_SIZE, sizeof(int));
    
    for (i = 0; i <= MAXJSAMPLE; i++) {
        rgb_ycc_tab[i+R_Y_OFF] = FIX(0.29900) * i;
        rgb_ycc_tab[i+G_Y_OFF] = FIX(0.58700) * i;
        rgb_ycc_tab[i+B_Y_OFF] = FIX(0.11400) * i     + ONE_HALF;
        rgb_ycc_tab[i+R_CB_OFF] = (-FIX(0.16874)) * i;
        rgb_ycc_tab[i+G_CB_OFF] = (-FIX(0.33126)) * i;
        /* We use a rounding fudge-factor of 0.5-epsilon for Cb and Cr.
         * This ensures that the maximum output will round to MAXJSAMPLE
         * not MAXJSAMPLE+1, and thus that we don't have to range-limit.
         */
        rgb_ycc_tab[i+B_CB_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
        /*  B=>Cb and R=>Cr tables are the same
            rgb_ycc_tab[i+R_CR_OFF] = FIX(0.50000) * i    + CBCR_OFFSET + ONE_HALF-1;
        */
        rgb_ycc_tab[i+G_CR_OFF] = (-FIX(0.41869)) * i;
        rgb_ycc_tab[i+B_CR_OFF] = (-FIX(0.08131)) * i;
    }

    cconvert->rgb_ycc_tab = rgb_ycc_tab;

    /*
     * assign methods
     */
    cconvert->pub.color_convert = bgr_ycc_convert; /* ToDo: info if rgb or bgr */
    
    cinfo->cconvert = (struct j3d_c_color_converter *) cconvert;
}


