/* 
 * cf3ddct.c: forward 3D-DCT object
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */

#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include "j3dlib.h"


#define CENTERJSAMPLE	128


/* Private subobject */
typedef struct {
    struct j3d_c_forward_3ddct pub; /* public fields */

    /* private variables */
    float divisors[DCT_MAXSIZE3];
    
} my_f3ddct;
typedef my_f3ddct * my_f3ddct_ptr;


#include "f3ddct.c"



/*
 * transform/quantize a DCTSIZE*DCTSIZE*depth block
 *
 * input:  imagesequence, offset
 * output: coefcube
 *
 */
static void forward_3ddct (j3d_compress_ptr cinfo, j3d_image_sequence_ptr seq, j3d_dct_coefcube coef, unsigned int start_off, unsigned int color_comp, float *divisors, j3d_quant_tbl quant_tbl)
/* This version is used for floating-point DCT implementations. */
{
    /* This routine is heavily used, so it's worth coding it tightly. */
    unsigned int depth = seq->depth;
    float workspace[DCT_MAXSIZE3]; /* work area for FDCT subroutine */

    
    /* Load data into workspace, applying unsigned->signed conversion */
    { register float *workspaceptr;
    register j3d_sample *elemptr;
    register unsigned int elemd;
    register unsigned int elemr;
    register unsigned int elem_next_row = seq->width - DCTSIZE;
    
    /* ToDo!: This is a big ugly hack for now, but seq->width should be an array[comp]! */ 
    elem_next_row = (color_comp == 0) ? seq->width - DCTSIZE : (seq->width >> 1) - DCTSIZE;
    

    workspaceptr = workspace;
    for(elemd = 0; elemd < depth; elemd++) {
        elemptr = &(seq->images[elemd][color_comp][start_off]);
        
        for (elemr = 0; elemr < DCTSIZE; elemr++) {
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);
            *workspaceptr++ = (float)(*elemptr++ - CENTERJSAMPLE);

            elemptr += elem_next_row;  /*  set ptr to next row  */
        }
    }
    

    /* Perform the DCT */
    j3d_f3ddct(workspace, depth);

    /* Quantize/descale the coefficients, and store into coef_blocks[] */
    { register float temp;
    register unsigned int i;
//    register j3d_dct_coefcube *output_ptr = coef;

    for (i = 0; i < DCTSIZE2 * depth; i++) {
	/* Apply the quantization and scaling factor */
	temp = workspace[i] * divisors[i] / ((float) quant_tbl[i]); /* ToDo: precalc div * quant */
	/* Round to nearest integer.
	 * Since C does not specify the direction of rounding for negative
	 * quotients, we have to force the dividend positive for portability.
	 * The maximum coefficient size is +-16K (for 12-bit data), so this
	 * code should work for either 16-bit or 32-bit ints.
	 */
	coef[i] = (j3d_dct_coef) ((int) (temp + (float) 16384.5) - 16384);
    }
    }
    }
}




static void f3ddct_start(j3d_compress_ptr cinfo, j3d_image_sequence_ptr in_seq, j3d_dct_coef_ptr out_coef) {
    int start_off;
    unsigned int color_comp;
    int cube_nr;
    int width, height;
    int x, y;
    my_f3ddct_ptr f3ddct = (my_f3ddct_ptr) cinfo->f3ddct;


    DEBUGP("starting 3DDCT transform\n");    


    /* process all blocks */
    for(color_comp = 0; color_comp < cinfo->color_components; color_comp++) {
        width = (in_seq->width / DCTSIZE) / cinfo->downsampling_factor[color_comp];
        height = (in_seq->height / DCTSIZE) / cinfo->downsampling_factor[color_comp];
        start_off = 0;
	cube_nr = 0;
    
        for(y = height; y > 0; y--) {
	    for(x = width; x > 0; x--) {

                forward_3ddct (cinfo, in_seq, (out_coef->color_component[color_comp])[cube_nr], start_off, color_comp, f3ddct->divisors, cinfo->quant_tbl_ptrs[cinfo->quant_tbl_no[color_comp]]);
                start_off += DCTSIZE;
                cube_nr++;
            }
            start_off += (DCTSIZE - 1) * (width * DCTSIZE);
        }
	out_coef->length[color_comp] = cube_nr;
        out_coef->width[color_comp] = width;
	DEBUGP("  component: %i, cubes: %i\n", color_comp, cube_nr);
    }



    out_coef->depth = in_seq->depth;


#if 0
#ifdef J3D_LOWDEBUG
    /*
     * show debug info about DC values of all cubes
     */
    {
        int i;
        DEBUGP("\n\nDC Coefficients:\n----------------\n\n");
        for(i = 0; i < out_coef->length[0]; i++) {
            DEBUGP("cube %i: DC = %i\n", i, out_coef->color_component[1][i][0]);
        }
        
    }
    
#endif
#endif

#ifdef J3D_LOWDEBUG
    /*
     * show debug info about the image and the cube
     */
    {
        int x,y,z,i;
        const int cube_nr = 10;
        const int img_off = cube_nr*DCTSIZE +  DCTSIZE * (cube_nr * DCTSIZE % in_seq->width);
        
        DEBUGP("\n\nImage Data-Cube %i:\n----------------\n\n", cube_nr);
        for(z = 0; z < 8; z++) {
            for(y = 0; y < 8; y++) {
                for(x = 0; x < 8; x++) {
                    DEBUGP("%i\t", in_seq->images[z][0][img_off + x + y * in_seq->width]);
                }
                DEBUGP("\n");
            }
            DEBUGP("\n");
        }

        DEBUGP("\n\nCoefficient Data-Cube %i:\n-------------------------\n\n", cube_nr);
        i = 0;
        for(z = 0; z < 8; z++) {
            for(y = 0; y < 8; y++) {
                for(x = 0; x < 8; x++) {
                    DEBUGP("%i\t", out_coef->color_component[1][cube_nr][i++]);
                }
                DEBUGP("\n");
            }
            DEBUGP("\n");
        }
    }
#endif

    
}




void j3d_init_forward_3ddct(j3d_compress_ptr cinfo) {
    my_f3ddct_ptr f3ddct;
    int i;
    
    f3ddct = (my_f3ddct_ptr) malloc(sizeof(my_f3ddct));
    
    
    /* init the objects used */


    /* init the divisors (quantisation) table */
    for(i = 0; i < DCT_MAXSIZE3; i++) {
      f3ddct->divisors[i] = (float) 4.0 / (256.0 * 0.707106796);
    }
    


    
    /*
     * assign methods
     */
    f3ddct->pub.transform = f3ddct_start;

    
    cinfo->f3ddct = (struct j3d_c_forward_3ddct *) f3ddct;
}

    
