/* 
 * cmarker.c: marker writer module
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */

#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include "j3dlib.h"


typedef enum {			/* JPEG marker codes */
    M_SOF0  = 0xc0,
    M_SOF1  = 0xc1,
    M_SOF2  = 0xc2,
    M_SOF3  = 0xc3,
  
    M_SOF5  = 0xc5,
    M_SOF6  = 0xc6,
    M_SOF7  = 0xc7,
  
    M_JPG   = 0xc8,
    M_SOF9  = 0xc9,
    M_SOF10 = 0xca,
    M_SOF11 = 0xcb,
  
    M_SOF13 = 0xcd,
    M_SOF14 = 0xce,
    M_SOF15 = 0xcf,
  
    M_DHT   = 0xc4,
  
    M_DAC   = 0xcc,
  
    M_RST0  = 0xd0,
    M_RST1  = 0xd1,
    M_RST2  = 0xd2,
    M_RST3  = 0xd3,
    M_RST4  = 0xd4,
    M_RST5  = 0xd5,
    M_RST6  = 0xd6,
    M_RST7  = 0xd7,
  
    M_SOI   = 0xd8,
    M_EOI   = 0xd9,
    M_SOS   = 0xda,
    M_DQT   = 0xdb,
    M_DNL   = 0xdc,
    M_DRI   = 0xdd,
    M_DHP   = 0xde,
    M_EXP   = 0xdf,
  
    M_APP0  = 0xe0,
    M_APP1  = 0xe1,
    M_APP2  = 0xe2,
    M_APP3  = 0xe3,
    M_APP4  = 0xe4,
    M_APP5  = 0xe5,
    M_APP6  = 0xe6,
    M_APP7  = 0xe7,
    M_APP8  = 0xe8,
    M_APP9  = 0xe9,
    M_APP10 = 0xea,
    M_APP11 = 0xeb,
    M_APP12 = 0xec,
    M_APP13 = 0xed,
    M_APP14 = 0xee,
    M_APP15 = 0xef,
  
    M_JPG0  = 0xf0,
    M_JPG13 = 0xfd,
    M_COM   = 0xfe,
  
    M_TEM   = 0x01,
  
    M_ERROR = 0x100
} J3D_MARKER;


/* Private state */

typedef struct {
    struct j3d_c_marker pub; /* public fields */

  /* state vars */
    unsigned int last_restart_interval; /* last DRI value emitted; 0 after SOI */
    
    unsigned int next_restart_num;
} my_marker;

typedef my_marker * my_marker_ptr;


/*
 * Basic output routines.
 *
 * Note that we do not support suspension while writing a marker.
 * Therefore, an application using suspension must ensure that there is
 * enough buffer space for the initial markers (typ. 600-700 bytes) before
 * calling jpeg_start_compress, and enough space to write the trailing EOI
 * (a few bytes) before calling jpeg_finish_compress.  Multipass compression
 * modes are not supported at all with suspension, so those two are the only
 * points where markers will be written.
 */
static void emit_byte (j3d_bitarray_ptr bitarray, int val)
/* Emit a byte */
{
    *(bitarray)->next_output_byte++ = (unsigned char) val;
    bitarray->bits_in_buffer += 8;
    if (--(bitarray)->free_in_buffer == 0) {
        /* ToDo: handle this */
        ERREXIT(-1, "marker->emit_byte: buffer empty. This case is not handled yet");
    }
}



static void emit_marker (j3d_bitarray_ptr bitarray, J3D_MARKER mark)
/* Emit a marker code */
{
    emit_byte(bitarray, 0xFF);
    emit_byte(bitarray, (int) mark);
}



static void emit_2bytes (j3d_bitarray_ptr bitarray, int value)
/* Emit a 2-byte integer; these are always MSB first in JPEG files */
{
    emit_byte(bitarray, (value >> 8) & 0xFF);
    emit_byte(bitarray, value & 0xFF);
}


/*
 * Routines to write specific marker types.
 */
static int emit_dqt (j3d_compress_ptr cinfo, int index)
/* Emit a DQT marker */
/* Returns the precision used (0 = 8bits, 1 = 16bits) for baseline checking */
{
    j3d_quant_tbl qtbl = cinfo->quant_tbl_ptrs[index];
    j3d_bitarray_ptr bitarray = cinfo->output->bitarray;
    int prec;
    int i;
    const int qtbl_size = DCTSIZE2 * cinfo->max_depth;
    unsigned int * zigzag_order;

    if (qtbl == NULL)
        ERREXIT(-1, "emit_dqt: cannot emit empty table\n");

    if (! cinfo->quant_tbl_sent[index]) {
        DEBUGP("emit DQT%i\n", index);

        if(cinfo->max_depth == 8) zigzag_order = cinfo->coef_zigzag_order_8;
        else ERREXIT(-1, "emit_dqt: max_depth != 8 not supported yet\n");
        
        prec = 0;
        for (i = 0; i < qtbl_size; i++) {
            if (qtbl[i] > 255)
                prec = 1;
        }

                              
        emit_marker(bitarray, M_DQT);

        emit_2bytes(bitarray, prec ? qtbl_size*2 + 1 + 2 : qtbl_size + 1 + 2);

        emit_byte(bitarray, index + (prec<<4));

        for (i = 0; i < qtbl_size; i++) {
            /* The table entries must be emitted in zigzag order. */
            unsigned int qval = qtbl[zigzag_order[i]];
            if (prec)
                emit_byte(bitarray, (int) (qval >> 8));
            emit_byte(bitarray, (int) (qval & 0xFF));
        }

        cinfo->quant_tbl_sent[index] = -1;
    }

    return prec;
}



static void emit_dht (j3d_compress_ptr cinfo, int index, int is_ac)
/* Emit a DHT marker */
{
    j3d_bitarray_ptr bitarray = cinfo->output->bitarray;
    j3d_huff_tbl * htbl;
    int length, i;
    int out_index;
    
    if (is_ac) {
        htbl = cinfo->ac_huff_tbl_ptrs[index];
        out_index = index + 0x10;		/* output index has AC bit set */
        is_ac = 1;
    } else {
        htbl = cinfo->dc_huff_tbl_ptrs[index];
        out_index = index;
    }

    if (htbl == NULL)
        ERREXIT(-1, "emit_dht: cannot emit empty table\n");


    
    if (! cinfo->huff_tbl_sent[is_ac][index]) {
        DEBUGP("emit DHT%i_%s\n", index, is_ac ? "ac" : "dc");

        emit_marker(bitarray, M_DHT);
    
        length = 0;
        for (i = 1; i <= 16; i++)
            length += htbl->bits[i];


        emit_2bytes(bitarray, length + 2 + 1 + 16);
        emit_byte(bitarray, out_index);

        for (i = 1; i <= 16; i++) {
            emit_byte(bitarray, htbl->bits[i]);
        }
        
        for (i = 0; i < length; i++) {
            emit_byte(bitarray, htbl->huffval[i]);
        }
        
        cinfo->huff_tbl_sent[is_ac][index] = -1;
    }
}


static void emit_dri (j3d_compress_ptr cinfo)
/* Emit a DRI marker */
{
    j3d_bitarray_ptr bitarray = cinfo->output->bitarray;

    DEBUGP("emit DRI\n");

    emit_marker(bitarray, M_DRI);
  
    emit_2bytes(bitarray, 4);	/* fixed length */

    emit_2bytes(bitarray, (int) cinfo->restart_interval);
}


static void emit_sof (j3d_compress_ptr cinfo, J3D_MARKER code)
/* Emit a SOF marker */
{
    j3d_bitarray_ptr bitarray = cinfo->output->bitarray;
    int ci;

    DEBUGP("emit SOF\n");

    emit_marker(bitarray, code);
  
    emit_2bytes(bitarray, 3 * cinfo->color_components + 2 + 5 + 1); /* length */

    /* Make sure image isn't bigger than SOF field can handle */
    if ((long) cinfo->height > 65535L || (long) cinfo->width > 65535L) {
        ERREXIT(-1, "SOF: image too big\n");
    }
    

    emit_byte(bitarray, BITS_IN_J3D_SAMPLE);
    emit_2bytes(bitarray, (int) cinfo->height);
    emit_2bytes(bitarray, (int) cinfo->width);

    emit_byte(bitarray, cinfo->color_components);

    for (ci = 0; ci < cinfo->color_components; ci++) {
        emit_byte(bitarray, ci);
        emit_byte(bitarray, (cinfo->downsampling_factor[ci] << 4) + cinfo->downsampling_factor[ci]);
        emit_byte(bitarray, cinfo->quant_tbl_no[ci]);
    }
}


static void emit_sos (j3d_compress_ptr cinfo)
/* Emit a SOS marker */
{
    my_marker_ptr marker = (my_marker_ptr) cinfo->marker;
    j3d_bitarray_ptr bitarray = cinfo->output->bitarray;
    int i, td, ta;
  
    DEBUGP("emit SOS\n");

    emit_marker(bitarray, M_SOS);
  
    emit_2bytes(bitarray, 2 * cinfo->color_components + 2 + 1 + 3); /* length */
  
    emit_byte(bitarray, cinfo->color_components);
  
    for (i = 0; i < cinfo->color_components; i++) {
        /* ToDo: add component ID */
        emit_byte(bitarray, i);   /* emit component-ID: make it easy, ID=comp_no */
        td = cinfo->dc_huff_tbl_no[i];
        ta = cinfo->ac_huff_tbl_no[i];
        emit_byte(bitarray, (td << 4) + ta);
    }

    emit_byte(bitarray, 0);  /* Ss, not used */
    emit_byte(bitarray, 0);  /* Se, not used */
    emit_byte(bitarray, (0 << 4) + 0); /* Ah, Al: not used */

    /* reset restart marker */
    marker->next_restart_num = 0;
}


/*
 * Write datastream header.
 * This consists of an SOI and optional APPn markers.
 * We recommend use of the JFIF marker, but not the Adobe marker,
 * when using YCbCr or grayscale data.  The JFIF marker should NOT
 * be used for any other JPEG colorspace.  The Adobe marker is helpful
 * to distinguish RGB, CMYK, and YCCK colorspaces.
 * Note that an application can write additional header markers after
 * jpeg_start_compress returns.
 */
static void write_file_header (j3d_compress_ptr cinfo)
{
    my_marker_ptr marker = (my_marker_ptr) cinfo->marker;

    DEBUGP("emit SOI\n");
    emit_marker(cinfo->output->bitarray, M_SOI);	/* first the SOI */

    /* SOI is defined to reset restart interval to 0 */
    marker->last_restart_interval = 0;
}



/*
 * Write frame header.
 * This consists of DQT and SOFn markers.
 * Note that we do not emit the SOF until we have emitted the DQT(s).
 * This avoids compatibility problems with incorrect implementations that
 * try to error-check the quant table numbers as soon as they see the SOF.
 */
static void write_frame_header (j3d_compress_ptr cinfo)
{
    int ci, prec;
  
    /* Emit DQT for each quantization table.
     * Note that emit_dqt() suppresses any duplicate tables.
     */
    prec = 0;
    for (ci = 0; ci < cinfo->color_components; ci++) {
        prec += emit_dqt(cinfo, cinfo->quant_tbl_no[ci]);
    }
    /* now prec is nonzero iff there are any 16-bit quant tables. */
  
    emit_sof(cinfo, M_SOF0);	/* SOF code for baseline implementation */
}


/*
 * Write scan header.
 * This consists of DHT or DAC markers, optional DRI, and SOS.
 * Compressed data will be written following the SOS.
 */
static void write_scan_header (j3d_compress_ptr cinfo)
{
    my_marker_ptr marker = (my_marker_ptr) cinfo->marker;
    int i;
    /* Emit Huffman tables.
     * Note that emit_dht() suppresses any duplicate tables.
     */
    for (i = 0; i < cinfo->color_components; i++) {
        /* Sequential mode: need both DC and AC tables */
        emit_dht(cinfo, cinfo->dc_huff_tbl_no[i], 0);
        emit_dht(cinfo, cinfo->ac_huff_tbl_no[i], -1);
    }

    /* Emit DRI if required --- note that DRI value could change for each scan.
     * We avoid wasting space with unnecessary DRIs, however.
     */
    if (cinfo->restart_interval != marker->last_restart_interval) {
        emit_dri(cinfo);
        marker->last_restart_interval = cinfo->restart_interval;
    }

    emit_sos(cinfo);
}


/*
 * Write datastream trailer.
 */
static void write_file_trailer (j3d_compress_ptr cinfo)
{
    DEBUGP("emit EOI\n");
    emit_marker(cinfo->output->bitarray, M_EOI);
}


/*
 * write correct restart marker
 */
static void write_restart_marker(j3d_compress_ptr cinfo)
{
    my_marker_ptr marker = (my_marker_ptr) cinfo->marker;

    DEBUGP("emit RST%i\n", marker->next_restart_num);
    
    emit_byte(cinfo->output->bitarray, 0xFF);
    emit_byte(cinfo->output->bitarray, M_RST0 + marker->next_restart_num);

    /* Update restart-interval state */
    marker->next_restart_num++;
    marker->next_restart_num &= 7;
}



/*
 * Initialize the marker writer module.
 */
void j3d_init_c_marker (j3d_compress_ptr cinfo)
{
    my_marker_ptr marker;
    int i;

    /* Create the subobject */
    marker = (my_marker_ptr) malloc(sizeof(my_marker));

    cinfo->marker = (struct j3d_c_marker *) marker;

    /* mark all tables as not sent */
    for(i = 0; i < MAX_QUANT_TBLS; i++) cinfo->quant_tbl_sent[i] = 0;
    for(i = 0; i < MAX_HUFF_TBLS; i++) {
        cinfo->huff_tbl_sent[0][i] = 0; /* dc_tbl */
        cinfo->huff_tbl_sent[1][i] = 0; /* ac_tbl */
    }
    

    
    /* Initialize method pointers */
    marker->pub.write_file_header = write_file_header;
    marker->pub.write_frame_header = write_frame_header;
    marker->pub.write_scan_header = write_scan_header;
    marker->pub.write_file_trailer = write_file_trailer;
    marker->pub.write_restart_marker = write_restart_marker;
    //  marker->pub.write_tables_only = write_tables_only;
    //  marker->pub.write_marker_header = write_marker_header;
    //  marker->pub.write_marker_byte = write_marker_byte;

    /* Initialize private state */
    marker->last_restart_interval = 0;
}
