/* 
 * cparam.c: paramer setting routines
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */


#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include "j3dlib.h"



#define IMAGE_COLOR_COMPONENTS 3
#define IMAGE_COMPONENT_Y  0
#define IMAGE_COMPONENT_Cr 1
#define IMAGE_COMPONENT_Cb 2

#define IMAGE_DOWN_FACTOR_Y 1
#define IMAGE_DOWN_FACTOR_Cr 2
#define IMAGE_DOWN_FACTOR_Cb 2

#define RESTART_INTERVAL 16



#define DCT_QUALITY 50


#define DCTDEPTH     8   /* The basic DCT block is 8x8x8 samples */



static void std_huff_tables (j3d_compress_ptr cinfo)
/* Set up the standard Huffman tables (cf. JPEG standard section K.3) */
/* IMPORTANT: these are only valid for 8-bit data precision! */
{
    static const unsigned char bits_dc_luminance[17] =
    { /* 0-base */ 0, 0, 1, 5, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0 };
    static const unsigned char val_dc_luminance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
    static const unsigned char bits_dc_chrominance[17] =
    { /* 0-base */ 0, 0, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0 };
    static const unsigned char val_dc_chrominance[] =
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
  
    static const unsigned char bits_ac_luminance[17] =
    { /* 0-base */ 0, 0, 2, 1, 3, 3, 2, 4, 3, 5, 5, 4, 4, 0, 0, 1, 0x7d };
    static const unsigned char val_ac_luminance[] =
    { 0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
      0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
      0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
      0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
      0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
      0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
      0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
      0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
      0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
      0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
      0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
      0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
      0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
      0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
      0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
      0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
      0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
      0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
      0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
      0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };
  
    static const unsigned char bits_ac_chrominance[17] =
    { /* 0-base */ 0, 0, 2, 1, 2, 4, 4, 3, 4, 7, 5, 4, 4, 0, 1, 2, 0x77 };
    static const unsigned char val_ac_chrominance[] =
    { 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21,
      0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
      0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91,
      0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
      0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34,
      0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26,
      0x27, 0x28, 0x29, 0x2a, 0x35, 0x36, 0x37, 0x38,
      0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48,
      0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
      0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
      0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
      0x79, 0x7a, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
      0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96,
      0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
      0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4,
      0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
      0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2,
      0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
      0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
      0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
      0xf9, 0xfa };
  
    j3d_add_huff_table(&cinfo->dc_huff_tbl_ptrs[0], bits_dc_luminance, val_dc_luminance);
    j3d_add_huff_table(&cinfo->ac_huff_tbl_ptrs[0], bits_ac_luminance, val_ac_luminance);
    j3d_add_huff_table(&cinfo->dc_huff_tbl_ptrs[1], bits_dc_chrominance, val_dc_chrominance);
    j3d_add_huff_table(&cinfo->ac_huff_tbl_ptrs[1], bits_ac_chrominance, val_ac_chrominance);
}




/*
 * Quantization table setup routines
 */
static void j3d_add_quant_table (j3d_compress_ptr cinfo, int which_tbl,
		      const unsigned int *basic_table,
		      int scale_factor, int force_baseline)
/* Define a quantization table equal to the basic_table times
 * a scale factor (given as a percentage).
 * If force_baseline is TRUE, the computed quantization table entries
 * are limited to 1..255 for JPEG baseline compatibility.
 */
{
  j3d_quant_tbl qtbl;
  int i;
  long temp;


  if (cinfo->quant_tbl_ptrs[which_tbl] == NULL)
    cinfo->quant_tbl_ptrs[which_tbl] = (j3d_quant_tbl) calloc(DCT_MAXSIZE3, sizeof(unsigned int));

  qtbl = cinfo->quant_tbl_ptrs[which_tbl];

  for (i = 0; i < DCT_MAXSIZE3; i++) {
    temp = ((long) basic_table[i] * scale_factor + 50L) / 100L;
    /* limit the values to the valid range */
    if (temp <= 0L) temp = 1L;
    if (temp > 32767L) temp = 32767L; /* max quantizer needed for 12 bits */
    if (force_baseline && temp > 255L)
      temp = 255L;		/* limit to baseline range if requested */
    qtbl[i] = (unsigned int) temp;
  }


#ifdef J3D_LOWDEBUG
    {
        int x,y,z,i;
        
 
        DEBUGP("\n\nQuantization Table %i:\n-------------------------\n\n", which_tbl);
        i = 0;
        for(z = 0; z < 8; z++) {
            for(y = 0; y < 8; y++) {
                for(x = 0; x < 8; x++) {
                    DEBUGP("%i\t", qtbl[i++]);
                }
                DEBUGP("\n");
            }
            DEBUGP("\n");
        }
    }
#endif
}



static void j3d_set_linear_quality (j3d_compress_ptr cinfo, int depth, int scale_factor, int force_baseline)
/* Set or change the 'quality' (quantization) setting, using default tables
 * and a straight percentage-scaling quality scale.  In most cases it's better
 * to use jpeg_set_quality (below); this entry point is provided for
 * applications that insist on a linear percentage scaling.
 */
{
    int i;
    
#if 0
  static const double depth_factor = 1.3;
  double factor;
  unsigned int qtbl[DCT_MAXSIZE3];
  int i, z;
#endif
  /* These are the sample quantization tables given in JPEG spec section K.1.
   * The spec says that the values given produce "good" quality, and
   * when divided by 2, "very good" quality.
   */
static const unsigned int luminance_experimental_tbl[DCT_MAXSIZE3] = {
1, 1, 4, 9, 16, 25, 36, 49, 
1, 1, 4, 9, 16, 25, 36, 49, 
4, 4, 5, 9, 16, 25, 36, 49, 
9, 9, 9, 12, 18, 26, 37, 49, 
16, 16, 16, 18, 22, 29, 39, 51, 
25, 25, 25, 26, 29, 35, 43, 55, 
36, 36, 36, 37, 39, 43, 50, 60, 
49, 49, 49, 49, 51, 55, 60, 69, 

1, 1, 4, 9, 16, 25, 36, 49, 
1, 2, 4, 9, 16, 25, 36, 49, 
4, 4, 5, 9, 16, 25, 36, 49, 
9, 9, 9, 12, 18, 26, 37, 49, 
16, 16, 16, 18, 22, 29, 39, 51, 
25, 25, 25, 26, 29, 35, 43, 55, 
36, 36, 36, 37, 39, 43, 50, 60, 
49, 49, 49, 49, 51, 55, 60, 69, 

4, 4, 5, 9, 16, 25, 36, 49, 
4, 4, 5, 9, 16, 25, 36, 49, 
5, 5, 7, 10, 17, 25, 36, 49, 
9, 9, 10, 13, 18, 26, 37, 49, 
16, 16, 17, 18, 23, 29, 39, 51, 
25, 25, 25, 26, 29, 35, 44, 55, 
36, 36, 36, 37, 39, 44, 51, 60, 
49, 49, 49, 49, 51, 55, 60, 69, 

9, 9, 9, 12, 18, 26, 37, 49, 
9, 9, 9, 12, 18, 26, 37, 49, 
9, 9, 10, 13, 18, 26, 37, 49, 
12, 12, 13, 15, 20, 28, 38, 50, 
18, 18, 18, 20, 24, 31, 40, 52, 
26, 26, 26, 28, 31, 36, 44, 55, 
37, 37, 37, 38, 40, 44, 51, 61, 
49, 49, 49, 50, 52, 55, 61, 69, 

16, 16, 16, 18, 22, 29, 39, 51, 
16, 16, 16, 18, 22, 29, 39, 51, 
16, 16, 17, 18, 23, 29, 39, 51, 
18, 18, 18, 20, 24, 31, 40, 52, 
22, 22, 23, 24, 27, 33, 42, 53, 
29, 29, 29, 31, 33, 38, 46, 57, 
39, 39, 39, 40, 42, 46, 53, 62, 
51, 51, 51, 52, 53, 57, 62, 71, 

25, 25, 25, 26, 29, 35, 43, 55, 
25, 25, 25, 26, 29, 35, 43, 55, 
25, 25, 25, 26, 29, 35, 44, 55, 
26, 26, 26, 28, 31, 36, 44, 55, 
29, 29, 29, 31, 33, 38, 46, 57, 
35, 35, 35, 36, 38, 43, 50, 60, 
43, 43, 44, 44, 46, 50, 56, 65, 
55, 55, 55, 55, 57, 60, 65, 73, 

36, 36, 36, 37, 39, 43, 50, 60, 
36, 36, 36, 37, 39, 43, 50, 60, 
36, 36, 36, 37, 39, 44, 51, 60, 
37, 37, 37, 38, 40, 44, 51, 61, 
39, 39, 39, 40, 42, 46, 53, 62, 
43, 43, 44, 44, 46, 50, 56, 65, 
50, 50, 51, 51, 53, 56, 62, 70, 
60, 60, 60, 61, 62, 65, 70, 78, 

49, 49, 49, 49, 51, 55, 60, 69, 
49, 49, 49, 49, 51, 55, 60, 69, 
49, 49, 49, 49, 51, 55, 60, 69, 
49, 49, 49, 50, 52, 55, 61, 69, 
51, 51, 51, 52, 53, 57, 62, 71, 
55, 55, 55, 55, 57, 60, 65, 73, 
60, 60, 60, 61, 62, 65, 70, 78, 
69, 69, 69, 69, 71, 73, 78, 84
  };
#if 0
  static const unsigned int std_luminance_quant_tbl[DCTSIZE2] = {
    16,  11,  10,  16,  24,  40,  51,  61,
    12,  12,  14,  19,  26,  58,  60,  55,
    14,  13,  16,  24,  40,  57,  69,  56,
    14,  17,  22,  29,  51,  87,  80,  62,
    18,  22,  37,  56,  68, 109, 103,  77,
    24,  35,  55,  64,  81, 104, 113,  92,
    49,  64,  78,  87, 103, 121, 120, 101,
    72,  92,  95,  98, 112, 100, 103,  99
  };

  static const unsigned int std_chrominance_quant_tbl[DCTSIZE2] = {
    17,  18,  24,  47,  99,  99,  99,  99,
    18,  21,  26,  66,  99,  99,  99,  99,
    24,  26,  56,  99,  99,  99,  99,  99,
    47,  66,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99
  };
#endif


  /* Set up two quantization tables using the specified scaling */
//  cinfo->quant_tbl_ptrs = (j3d_quant_tbl *) calloc(cinfo->color_components, sizeof(j3d_quant_tbl));

  /* zero all tables */
  for(i = 0; i < MAX_QUANT_TBLS; i++) cinfo->quant_tbl_ptrs[i] = NULL;
  

  /* This is a first approach to use the JPEG standard tables in J3D.
   * ToDo: considerations on quant tables for J3D
   */
#if 0
  factor = 1.0;
  for(z = 0; z < depth; z++) {
    for(i = 0; i < DCTSIZE2; i++) {
      qtbl[z * DCTSIZE2 + i] = (unsigned int) 
	((double) std_luminance_quant_tbl[i] * factor);
    } 
    factor *= depth_factor;
  }
  qtbl[0] = 1;
  j3d_add_quant_table(cinfo, 0, qtbl,
		       scale_factor, force_baseline);



  factor = 1.0;
  for(z = 0; z < depth; z++) {
    for(i = 0; i < DCTSIZE2; i++) {
      qtbl[z * DCTSIZE2 + i] = (unsigned int) 
	((double) std_chrominance_quant_tbl[i] * factor);
    } 
    factor *= depth_factor;
  }
  qtbl[0] = 1;
  j3d_add_quant_table(cinfo, 1, qtbl,
		       scale_factor, force_baseline);

  /* assign tables to color components */
  cinfo->quant_tbl_no[0] = 0;
  cinfo->quant_tbl_no[1] = 1;
  cinfo->quant_tbl_no[2] = 1;
#endif

  j3d_add_quant_table(cinfo, 0, luminance_experimental_tbl,
		       scale_factor, force_baseline);
  j3d_add_quant_table(cinfo, 1, luminance_experimental_tbl,
		       scale_factor, force_baseline);
  
  /* assign tables to color components */
  cinfo->quant_tbl_no[0] = 0;
  cinfo->quant_tbl_no[1] = 1;
  cinfo->quant_tbl_no[2] = 1;
  
}




static int j3d_quality_scaling (int quality)
/* Convert a user-specified quality rating to a percentage scaling factor
 * for an underlying quantization table, using our recommended scaling curve.
 * The input 'quality' factor should be 0 (terrible) to 100 (very good).
 */
{
  /* Safety limit on quality factor.  Convert 0 to 1 to avoid zero divide. */
  if (quality <= 0) quality = 1;
  if (quality > 100) quality = 100;

  /* The basic table is used as-is (scaling 100) for a quality of 50.
   * Qualities 50..100 are converted to scaling percentage 200 - 2*Q;
   * note that at Q=100 the scaling is 0, which will cause jpeg_add_quant_table
   * to make all the table entries 1 (hence, minimum quantization loss).
   * Qualities 1..50 are converted to scaling percentage 5000/Q.
   */
  if (quality < 50)
    quality = 5000 / quality;
  else
    quality = 200 - quality*2;

  return quality;
}




void j3d_set_quality (j3d_compress_ptr cinfo, int quality, int force_baseline)
/* Set or change the 'quality' (quantization) setting, using default tables.
 * This is the standard quality-adjusting entry point for typical user
 * interfaces; only those who want detailed control over quantization tables
 * would use the preceding three routines directly.
 */
{
  /* Convert user 0-100 rating to percentage scaling */
  quality = j3d_quality_scaling(quality);

  /* Set up standard quality tables */
  j3d_set_linear_quality(cinfo, DCTDEPTH, quality, force_baseline);
}



    

void j3d_c_set_std_params(j3d_compress_ptr cinfo) {
    /* set the basic components */
//    cinfo->width = IMAGE_WIDTH;
//    cinfo->height = IMAGE_HEIGHT;
    cinfo->max_depth = DCTDEPTH;

    cinfo->color_components = IMAGE_COLOR_COMPONENTS;

    cinfo->restart_interval = RESTART_INTERVAL;

    /* ToDo: this is a hack for now, we should use color tables etc. */
    cinfo->downsampling_factor[IMAGE_COMPONENT_Y] = IMAGE_DOWN_FACTOR_Y;
    cinfo->downsampling_factor[IMAGE_COMPONENT_Cr] = IMAGE_DOWN_FACTOR_Cr;
    cinfo->downsampling_factor[IMAGE_COMPONENT_Cb] = IMAGE_DOWN_FACTOR_Cb;


    cinfo->coef_zigzag_order_8 = j3d_std_coef_zigzag();


    /* init huffman tables */
    std_huff_tables(cinfo);
    cinfo->ac_huff_tbl_no[0] = 0;
    cinfo->ac_huff_tbl_no[1] = 1;
    cinfo->ac_huff_tbl_no[2] = 1;
    cinfo->dc_huff_tbl_no[0] = 0;
    cinfo->dc_huff_tbl_no[1] = 1;
    cinfo->dc_huff_tbl_no[2] = 1;
    

    
    /* init quantization tables */
    j3d_set_quality(cinfo, DCT_QUALITY, -1);
    
}


/* old quant stuff */
#if 0
static void set_std_quantization_table(j3d_compress_ptr cinfo, int comp)
{
    int x, y, z, i;
    float factor;
    j3d_quant_tbl qtbl;
    
    if(comp == 0) factor = 1.0;
    else factor = 4.0;
    

    qtbl = (float *)calloc(DCT_MAXSIZE3, sizeof(float));

    
    i = 0;
    for(z = 0; z < DCTSIZE; z++) {
        for(y = 0; y < DCTSIZE; y++) {
            for(x = 0; x < DCTSIZE; x++) {
                if(i == 0) qtbl[i++] = 1.0; /* DC value */
                else qtbl[i++] = 1.0 / (sqrt(1.0+x*x+y*y+z*z) * factor);
            }
        }
    }

    cinfo->quant_tbl[comp] = qtbl;


#ifdef j3D_LOWDEBUG
    {
        int x,y,z,i;
        
 
        DEBUGP("\n\nQuantization Table %i:\n-------------------------\n\n", comp
);
        i = 0;
        for(z = 0; z < 8; z++) {
            for(y = 0; y < 8; y++) {
                for(x = 0; x < 8; x++) {
                    DEBUGP("%f\t", qtbl[i++]);
                }
                DEBUGP("\n");
            }
            DEBUGP("\n");
        }
    }
#endif
}

#endif



#if 0
/* print experimental table */
int main()
{
    int x, y, z, i;
    float factor;
    int * qtbl;
    
    factor = 1.0;
    

    qtbl = (int *)calloc(8*8*8, sizeof(int));

    
    i = 0;
    for(z = 0; z < 8; z++) {
        for(y = 0; y < 8; y++) {
            for(x = 0; x < 8; x++) {
                if(i == 0) qtbl[i++] = 1.0; /* DC value */
                else qtbl[i++] = (sqrt(1.0+x*x*x+y*y*y+z*z) * factor);
            }
        }
    }


    {
        int x,y,z,i;
        
 
        i = 0;
        for(z = 0; z < 8; z++) {
            for(y = 0; y < 8; y++) {
                for(x = 0; x < 8; x++) {
                    printf("%i, ", qtbl[i++]);
                }
                printf("\n");
            }
            printf("\n");
        }
    }


    return 0;
}
#endif
