/* 
 * cvlc.c: variable length coder module
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */

/* 
 * ToDo: create new object bitarray_manager
 */


#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "j3dlib.h"


#define MAX_COEF_BITS 10




/* Derived data constructed for each Huffman table */

typedef struct {
    unsigned int ehufco[256];	/* code for each symbol */
    char ehufsi[256];		/* length of code for each symbol */
    /* If no code has been allocated for a symbol S, ehufsi[S] contains 0 */
} c_derived_tbl;



/* Private subobject */
typedef struct {
    struct j3d_c_vlc pub; /* public fields */

    /* private variables */
    c_derived_tbl * dc_derived_tbls[MAX_HUFF_TBLS];
    c_derived_tbl * ac_derived_tbls[MAX_HUFF_TBLS];
    
  /* state vars */
    int last_dc_val[MAX_COLOR_COMPONENTS]; /* last DC coef */
} my_vlc;
typedef my_vlc * my_vlc_ptr;



/* Outputting bytes to the file */

/* Emit a byte, taking 'action' if must suspend. */
#define emit_byte(bitarray,val,action)  \
	{ *(bitarray)->next_output_byte++ = (unsigned char) (val);  \
          bitarray->bits_in_buffer += 8; \
	  if (--(bitarray)->free_in_buffer == 0)  \
	      { action; } }




/* Outputting bits to the buffer */

/* Only the right 24 bits of put_buffer are used; the valid bits are
 * left-justified in this part.  At most 16 bits can be passed to emit_bits
 * in one call, and we never retain more than 7 bits in put_buffer
 * between calls, so 24 bits are sufficient.
 */

inline static int emit_bits (j3d_bitarray_ptr bitarray, unsigned int code, int size)
/* Emit some bits; return TRUE if successful, FALSE if must suspend */
{
    /* This routine is heavily used, so it's worth coding tightly. */
    register int32 put_buffer = (int32) code;
    register int put_bits = bitarray->put_bits;

    /* if size is 0, caller used an invalid Huffman table entry */
    if (size == 0)
        ERREXIT(-1, "vlc->emit_bits: invalid Huffman table entry\n");

    put_buffer &= (((int32) 1)<<size) - 1; /* mask off any extra bits in code */
  
    put_bits += size;		/* new number of bits in buffer */
  
    put_buffer <<= 24 - put_bits; /* align incoming bits */

    put_buffer |= bitarray->put_buffer; /* and merge with old buffer contents */
  
    while (put_bits >= 8) {
        int c = (int) ((put_buffer >> 16) & 0xFF);
    
        emit_byte(bitarray, c, return 0);
        if (c == 0xFF) {		/* need to stuff a zero byte? */
            emit_byte(bitarray, 0, return 0);
        }
        put_buffer <<= 8;
        put_bits -= 8;
    }

    bitarray->put_buffer = put_buffer; /* update bitarray variables */
    bitarray->put_bits = put_bits;

    return -1;
}


/* 
 * write the remaining bits in put_buffer (rest of byte is stuffed with zeroes)
 */
inline static void emit_remaining_bits(j3d_bitarray_ptr bitarray) {
    int remaining_bits = bitarray->put_bits;
    if(bitarray->put_bits > 0) {
        /* put zeroes, so one byte with remaining bits will be emitted */
        emit_bits(bitarray, 0, 8);
        bitarray->put_bits = 0;
        /* adjust bits_in_buffer */
        bitarray->bits_in_buffer -= 8 - remaining_bits;
    }
}



/* Encode a single cube's worth of coefficients */

static int encode_one_cube (j3d_bitarray_ptr bitarray, j3d_dct_coefcube cube, unsigned int depth, c_derived_tbl *dctbl, c_derived_tbl *actbl, int last_dc_val, unsigned int * coef_zigzag_tbl)
{
    register int temp, temp2;
    register int nbits;
    register unsigned int k; 
    register int r, i;
  
    /* Encode the DC coefficient difference per section F.1.2.1 */
  
    temp = temp2 = cube[0] - last_dc_val;
    
    if (temp < 0) {
        temp = -temp;		/* temp is abs value of input */
        /* For a negative input, want temp2 = bitwise complement of abs(input) */
        /* This code assumes we are on a two's complement machine */
        temp2--;
    }
  
    /* Find the number of bits needed for the magnitude of the coefficient */
    nbits = 0;
    while (temp) {
        nbits++;
        temp >>= 1;
    }
    /* Check for out-of-range coefficient values.
     * Since we're encoding a difference, the range limit is twice as much.
     */
    if (nbits > MAX_COEF_BITS+1)
        perror("!!! too many bits !!!\n");
  
//    ERREXIT(bitarray->cinfo, JERR_BAD_DCT_COEF);
  
    /* Emit the Huffman-coded symbol for the number of bits */
    if (! emit_bits(bitarray, dctbl->ehufco[nbits], dctbl->ehufsi[nbits]))
        return 0;

    /* Emit that number of bits of the value, if positive, */
    /* or the complement of its magnitude, if negative. */
    if (nbits)			/* emit_bits rejects calls with size 0 */
        if (! emit_bits(bitarray, (unsigned int) temp2, nbits))
            return 0;

    /* Encode the AC coefficients per section F.1.2.2 */
  
    r = 0;			/* r = run length of zeros */
  
    for (k = 1; k < DCTSIZE2*depth; k++) {
        if ((temp = cube[coef_zigzag_tbl[k]]) == 0) {
            r++;
        } else {
            /* if run length > 15, must emit special run-length-16 codes (0xF0) */
            while (r > 15) {
                if (! emit_bits(bitarray, actbl->ehufco[0xF0], actbl->ehufsi[0xF0]))
                    return 0;
                r -= 16;
            }

            temp2 = temp;
            if (temp < 0) {
                temp = -temp;		/* temp is abs value of input */
                /* This code assumes we are on a two's complement machine */
                temp2--;
            }
      
            /* Find the number of bits needed for the magnitude of the coefficient */
            nbits = 1;		/* there must be at least one 1 bit */
            while ((temp >>= 1))
                nbits++;
            /* Check for out-of-range coefficient values */
            if (nbits > MAX_COEF_BITS)
                fprintf(stderr, "!!!too many bits AC: %i !!!\n", nbits);
//                ERREXIT(bitarray->cinfo, JERR_BAD_DCT_COEF);
      
            /* Emit Huffman symbol for run length / number of bits */
            i = (r << 4) + nbits;
            if (! emit_bits(bitarray, actbl->ehufco[i], actbl->ehufsi[i]))
                return 0;

            /* Emit that number of bits of the value, if positive, */
            /* or the complement of its magnitude, if negative. */
            if (! emit_bits(bitarray, (unsigned int) temp2, nbits))
                return 0;
      
            r = 0;
        }
    }

    /* If the last coef(s) were zero, emit an end-of-cube code */
    if (r > 0)
        if (! emit_bits(bitarray, actbl->ehufco[0], actbl->ehufsi[0]))
            return 0;

    return -1;
}






/*
 * Compute the derived values for a Huffman table.
 * This routine also performs some validation checks on the table.
 *
 * ToDo: Error management
 */

static void j3d_make_c_derived_tbl (j3d_compress_ptr cinfo, int isDC, int tblno, c_derived_tbl ** pdtbl)
{
    j3d_huff_tbl *htbl;
    c_derived_tbl *dtbl;
    int p, i, l, lastp, si, maxsymbol;
    char huffsize[257];
    unsigned int huffcode[257];
    unsigned int code;

    /* Note that huffsize[] and huffcode[] are filled in code-length order,
     * paralleling the order of the symbols themselves in htbl->huffval[].
     */

    /* Find the input Huffman table */
//  if (tblno < 0 || tblno >= MAX_HUFF_TBLS)
//    ERREXIT1(cinfo, JERR_NO_HUFF_TABLE, tblno);
    htbl =
        isDC ? cinfo->dc_huff_tbl_ptrs[tblno] : cinfo->ac_huff_tbl_ptrs[tblno];
//  if (htbl == NULL)
//    ERREXIT1(cinfo, JERR_NO_HUFF_TABLE, tblno);

    /* Allocate a workspace if we haven't already done so. */
//    if (pdtbl[tblno] == NULL)
        pdtbl[tblno] = (c_derived_tbl *) malloc(sizeof(c_derived_tbl));
    dtbl = pdtbl[tblno];
  
    /* Figure C.1: make table of Huffman code length for each symbol */

    p = 0;
    for (l = 1; l <= 16; l++) {
        i = (int) htbl->bits[l];
//        if (i < 0 || p + i > 256)	/* protect against table overrun */
//            ERREXIT(cinfo, JERR_BAD_HUFF_TABLE);
        while (i--)
            huffsize[p++] = (char) l;
    }
    huffsize[p] = 0;
    lastp = p;
  
    /* Figure C.2: generate the codes themselves */
    /* We also validate that the counts represent a legal Huffman code tree. */

    code = 0;
    si = huffsize[0];
    p = 0;
    while (huffsize[p]) {
        while (((int) huffsize[p]) == si) {
            huffcode[p++] = code;
            code++;
        }
        /* code is now 1 more than the last code used for codelength si; but
         * it must still fit in si bits, since no code is allowed to be all ones.
         */
//    if (((int32) code) >= (((int32) 1) << si))
//      ERREXIT(cinfo, JERR_BAD_HUFF_TABLE);
        code <<= 1;
        si++;
    }
  
    /* Figure C.3: generate encoding tables */
    /* These are code and size indexed by symbol value */

    /* Set all codeless symbols to have code length 0;
     * this lets us detect duplicate VAL entries here, and later
     * allows emit_bits to detect any attempt to emit such symbols.
     */
    memset((void *)(dtbl->ehufsi), 0, (size_t)sizeof(dtbl->ehufsi));
    
      
    /* This is also a convenient place to check for out-of-range
     * and duplicated VAL entries.  We allow 0..255 for AC symbols
     * but only 0..15 for DC.  (We could constrain them further
     * based on data depth and mode, but this seems enough.)
     */
    maxsymbol = isDC ? 15 : 255;

    for (p = 0; p < lastp; p++) {
        i = htbl->huffval[p];
//    if (i < 0 || i > maxsymbol || dtbl->ehufsi[i])
//      ERREXIT(cinfo, JERR_BAD_HUFF_TABLE);
        dtbl->ehufco[i] = huffcode[p];
        dtbl->ehufsi[i] = huffsize[p];
    }
}







/*
 * Encode one MCU, write to bitarray
 * returns TRUE if last MCU in sequence encoded
 */
static int encode_mcu(j3d_compress_ptr cinfo, j3d_bitarray_ptr out_bitarray) {
    my_vlc_ptr vlc = (my_vlc_ptr) cinfo->vlc;
    j3d_dct_coefcube cube;
    c_derived_tbl *dctbl;
    c_derived_tbl *actbl;
    unsigned int curr_comp;
    unsigned int depth;
    int retval = 0;

//    DEBUGP("encoding MCU\n");
    
    
    
    while(!(retval & J3D_MCU_COMPLETE) && ((retval = cinfo->mcu->get_next_cube(cinfo, &cube, &curr_comp, &depth)) & J3D_MCU_OK)) {
//        DEBUGP(".");
        dctbl = vlc->dc_derived_tbls[cinfo->ac_huff_tbl_no[curr_comp]];
        actbl = vlc->ac_derived_tbls[cinfo->ac_huff_tbl_no[curr_comp]];

        /* encode cube */
        if(!encode_one_cube(out_bitarray, cube, depth, dctbl, actbl,
                            vlc->last_dc_val[curr_comp],
                            cinfo->coef_zigzag_order_8)) {
            perror("encoding cube\n"); /* should not happen (?) */
        }
        
        /* save DC value */
        vlc->last_dc_val[curr_comp] = cube[0];
    }
//    DEBUGP("\n");

    
    return (retval & J3D_MCU_SEQ_COMPLETE) || (retval & J3D_MCU_NO_MORE_BLOCKS);
}





/*
 * returns TRUE when all cubes are encoded
 */
static int vlc_start(j3d_compress_ptr cinfo){
    int interval_remain;
    int retval = 0;
    my_vlc_ptr vlc = (my_vlc_ptr) cinfo->vlc;
    j3d_bitarray_ptr bitarray = cinfo->output->bitarray;
    
    
    DEBUGP("encoding group of cubes\n");

    /* init state */

    vlc->last_dc_val[0] = 0;
    vlc->last_dc_val[1] = 0;
    vlc->last_dc_val[2] = 0;

    

//    DEBUGP("comp=%i, cube=%i, GOC=%i\n", vlc->curr_component, vlc->curr_cube, goc_remain);
    
    /* process GOC */
    interval_remain = cinfo->restart_interval;
    while(interval_remain > 0) { 
        if(encode_mcu(cinfo, bitarray)) { /* end of sequence (no more MCU's) */
            interval_remain = 0;
            retval = -1;
        }
        else interval_remain--;
    }
    
    
    /* here we want only full bytes */
    emit_remaining_bits(bitarray);
//    out_bitarray->bits_in_buffer = out_bitarray->bits_in_buffer;

    
    return retval;
}




void j3d_init_c_vlc(j3d_compress_ptr cinfo) {
    my_vlc_ptr vlc;
    
    vlc = (my_vlc_ptr) malloc(sizeof(my_vlc));

    /* ToDo: correct loop */
    j3d_make_c_derived_tbl(cinfo, -1, 0, vlc->dc_derived_tbls);
    j3d_make_c_derived_tbl(cinfo, -1, 1, vlc->dc_derived_tbls);
    j3d_make_c_derived_tbl(cinfo,  0, 0, vlc->ac_derived_tbls);
    j3d_make_c_derived_tbl(cinfo,  0, 1, vlc->ac_derived_tbls);


    /* init the objects used */


    /*
     * assign methods
     */
    vlc->pub.encode_goc = vlc_start;

    
    cinfo->vlc = (struct j3d_c_vlc *) vlc;
}

    
