/* 
 * dcolor.c: color conversion and upsampling routines.
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */

#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "j3dlib.h"




/* Private subobject */
typedef struct {
    struct j3d_d_color_deconverter pub;

    /* public fields */

    int * Cr_r_tab;		/* => table for Cr to R conversion */
    int * Cb_b_tab;		/* => table for Cb to B conversion */
    int32 * Cr_g_tab;		/* => table for Cr to G conversion */
    int32 * Cb_g_tab;		/* => table for Cb to G conversion */

    j3d_sample * range_limit_table;
} my_color_converter;
typedef my_color_converter * my_cconvert_ptr;



/**************** YCbCr -> RGB conversion: most common case **************/

/*
 * YCbCr is defined per CCIR 601-1, except that Cb and Cr are
 * normalized to the range 0..MAXJSAMPLE rather than -0.5 .. 0.5.
 * The conversion equations to be implemented are therefore
 *	R = Y                + 1.40200 * Cr
 *	G = Y - 0.34414 * Cb - 0.71414 * Cr
 *	B = Y + 1.77200 * Cb
 * where Cb and Cr represent the incoming values less CENTERJSAMPLE.
 * (These numbers are derived from TIFF 6.0 section 21, dated 3-June-92.)
 *
 * To avoid floating-point arithmetic, we represent the fractional constants
 * as integers scaled up by 2^16 (about 4 digits precision); we have to divide
 * the products by 2^16, with appropriate rounding, to get the correct answer.
 * Notice that Y, being an integral input, does not contribute any fraction
 * so it need not participate in the rounding.
 *
 * For even more speed, we avoid doing any multiplications in the inner loop
 * by precalculating the constants times Cb and Cr for all possible values.
 * For 8-bit JSAMPLEs this is very reasonable (only 256 entries per table);
 * for 12-bit samples it is still acceptable.  It's not very reasonable for
 * 16-bit samples, but if you want lossless storage you shouldn't be changing
 * colorspace anyway.
 * The Cr=>R and Cb=>B values can be rounded to integers in advance; the
 * values for the G calculation are left scaled up, since we must add them
 * together before rounding.
 */

#define MAXJSAMPLE      255
#define CENTERJSAMPLE	128

#define SCALEBITS	16	/* speediest right-shift on some machines */
#define CBCR_OFFSET	((int) CENTERJSAMPLE << SCALEBITS)
#define ONE_HALF	((int) 1 << (SCALEBITS-1))
#define FIX(x)		((int) ((x) * (1L<<SCALEBITS) + 0.5))



/*
 * Initialize tables for YCC->RGB colorspace conversion.
 */

static void build_ycc_rgb_table (my_cconvert_ptr cconvert) {
    int i;
    int32 x;

    cconvert->Cr_r_tab = (int *) calloc((MAXJSAMPLE+1), sizeof(int));
    cconvert->Cb_b_tab = (int *) calloc((MAXJSAMPLE+1), sizeof(int));
    cconvert->Cr_g_tab = (int32 *) calloc((MAXJSAMPLE+1), sizeof(int32));
    cconvert->Cb_g_tab = (int32 *) calloc((MAXJSAMPLE+1), sizeof(int32));

  
    for (i = 0, x = -CENTERJSAMPLE; i <= MAXJSAMPLE; i++, x++) {
        /* i is the actual input pixel value, in the range 0..MAXJSAMPLE */
        /* The Cb or Cr value we are thinking of is x = i - CENTERJSAMPLE */
        /* Cr=>R value is nearest int to 1.40200 * x */
        cconvert->Cr_r_tab[i] = (int) (FIX(1.40200) * x + ONE_HALF) >> SCALEBITS;
        /* Cb=>B value is nearest int to 1.77200 * x */
        cconvert->Cb_b_tab[i] = (int) (FIX(1.77200) * x + ONE_HALF) >> SCALEBITS;
        /* Cr=>G value is scaled-up -0.71414 * x */
        cconvert->Cr_g_tab[i] = (- FIX(0.71414)) * x;
        /* Cb=>G value is scaled-up -0.34414 * x */
        /* We also add in ONE_HALF so that need not do it in inner loop */
        cconvert->Cb_g_tab[i] = (- FIX(0.34414)) * x + ONE_HALF;
    }
}



/*
 * Several decompression processes need to range-limit values to the range
 * 0..MAXJSAMPLE; the input value may fall somewhat outside this range
 * due to noise introduced by quantization, roundoff error, etc.  These
 * processes are inner loops and need to be as fast as possible.  On most
 * machines, particularly CPUs with pipelines or instruction prefetch,
 * a (subscript-check-less) C table lookup
 *		x = sample_range_limit[x];
 * is faster than explicit tests
 *		if (x < 0)  x = 0;
 *		else if (x > MAXJSAMPLE)  x = MAXJSAMPLE;
 * These processes all use a common table prepared by the routine below.
 *
 * For most steps we can mathematically guarantee that the initial value
 * of x is within MAXJSAMPLE+1 of the legal range, so a table running from
 * -(MAXJSAMPLE+1) to 2*MAXJSAMPLE+1 is sufficient.  But for the initial
 * limiting step (just after the IDCT), a wildly out-of-range value is 
 * possible if the input data is corrupt.  To avoid any chance of indexing
 * off the end of memory and getting a bad-pointer trap, we perform the
 * post-IDCT limiting thus:
 *		x = range_limit[x & MASK];
 * where MASK is 2 bits wider than legal sample data, ie 10 bits for 8-bit
 * samples.  Under normal circumstances this is more than enough range and
 * a correct output will be generated; with bogus input data the mask will
 * cause wraparound, and we will safely generate a bogus-but-in-range output.
 * For the post-IDCT step, we want to convert the data from signed to unsigned
 * representation by adding CENTERJSAMPLE at the same time that we limit it.
 * So the post-IDCT limiting table ends up looking like this:
 *   CENTERJSAMPLE,CENTERJSAMPLE+1,...,MAXJSAMPLE,
 *   MAXJSAMPLE (repeat 2*(MAXJSAMPLE+1)-CENTERJSAMPLE times),
 *   0          (repeat 2*(MAXJSAMPLE+1)-CENTERJSAMPLE times),
 *   0,1,...,CENTERJSAMPLE-1
 * Negative inputs select values from the upper half of the table after
 * masking.
 *
 * We can save some space by overlapping the start of the post-IDCT table
 * with the simpler range limiting table.  The post-IDCT table begins at
 * sample_range_limit + CENTERJSAMPLE.
 *
 * Note that the table is allocated in near data space on PCs; it's small
 * enough and used often enough to justify this.
 */

static void build_range_limit_table (my_cconvert_ptr cconvert)
/* Allocate and fill in the sample_range_limit table */
{
    j3d_sample * table;
    int i;

    table = (j3d_sample *) calloc((5 * (MAXJSAMPLE+1) + CENTERJSAMPLE), sizeof(j3d_sample));
    table += (MAXJSAMPLE+1);	/* allow negative subscripts of simple table */
    cconvert->range_limit_table = table;
    
    /* First segment of "simple" table: limit[x] = 0 for x < 0 */
    memset(table - (MAXJSAMPLE+1), 0, (size_t) (MAXJSAMPLE+1) * sizeof(j3d_sample));
    /* Main part of "simple" table: limit[x] = x */
    for (i = 0; i <= MAXJSAMPLE; i++)
        table[i] = (j3d_sample) i;
    table += CENTERJSAMPLE;	/* Point to where post-IDCT table starts */
    /* End of simple table, rest of first half of post-IDCT table */
    for (i = CENTERJSAMPLE; i < 2*(MAXJSAMPLE+1); i++)
        table[i] = MAXJSAMPLE;
    /* Second half of post-IDCT table */
    memset(table + (2 * (MAXJSAMPLE+1)), 0,
           (size_t) (2 * (MAXJSAMPLE+1) - CENTERJSAMPLE) * sizeof(j3d_sample));
    memset(table + (4 * (MAXJSAMPLE+1) - CENTERJSAMPLE), MAXJSAMPLE,
           (size_t) CENTERJSAMPLE * sizeof(j3d_sample));
}



static void upsample_2(j3d_color_component buf, unsigned int width, unsigned int height) {
    register int x, y;
    register int inptr, outptr0, outptr1;
    register j3d_sample in;

    outptr0 = (width << 1) * ((height << 1) - 1);
    outptr1 = (width << 1) * ( height << 1     );
    inptr = width * height;
    
    for(y = 0; y < height; y++)  {
        for(x = 0; x < width; x++)  {
            in = buf[--inptr];
            buf[--outptr0] = in;
            buf[--outptr0] = in;
            buf[--outptr1] = in;
            buf[--outptr1] = in;
        }
        outptr0 -= width << 1;
        outptr1 -= width << 1;
    }  
}



/*
 * color upsampling and conversion.
 * ToDo: efficient algorithm which does both in one pass
 */
static void ycc_rgb_convert(j3d_decompress_ptr cinfo, j3d_image input_buf, j3d_rgb_image output_buf) {
    my_cconvert_ptr cconvert = (my_cconvert_ptr) cinfo->cconvert;
  
    register unsigned int incount, outcount;
    register int y, cb, cr;

    /* copy these pointers into registers if possible */
    register int * Crrtab = cconvert->Cr_r_tab;
    register int * Cbbtab = cconvert->Cb_b_tab;
    register int32 * Crgtab = cconvert->Cr_g_tab;
    register int32 * Cbgtab = cconvert->Cb_g_tab;
    register j3d_color_component comp0, comp1, comp2;
    register unsigned int imagesize = (cinfo->width * cinfo->height);
    register j3d_sample * range_limit = cconvert->range_limit_table;

//    register int tmp;
    
    
    comp0 = input_buf[0];
    comp1 = input_buf[1];
    comp2 = input_buf[2];
    
    
    /* color upsampling */
    /* ToDo: upsampling for all possible values of factor (see ccolor.c) */
    
    /* Y */
    
    /* Cb */
    upsample_2(comp1, cinfo->width >> 1, cinfo->height >> 1);

    
    /* Cr */
    upsample_2(comp2, cinfo->width >> 1, cinfo->height >> 1);


    
    /* color conversion */
    outcount = 0;
    
    
    for(incount = 0; incount < imagesize; incount++) {
        y  = comp0[incount];
        cb = comp1[incount];
        cr = comp2[incount];


        /* ToDo: range limit table is not big enough (I think...) there are still some errors */
        
/*          *output_buf++ = range_limit[y + Crrtab[cr]]; */
/*          *output_buf++ = range_limit[y + ((Cbgtab[cb] + Crgtab[cr]) >> SCALEBITS)]; */
/*          *output_buf++ = range_limit[y + Cbbtab[cb]]; */
        {
            int tmp;
            
            tmp = y + Crrtab[cr];
            if(tmp > 255) *output_buf++ = 255;
            else if(tmp < 0) *output_buf++ = 0;
            else *output_buf++ = tmp;
            
            tmp = y + ((Cbgtab[cb] + Crgtab[cr]) >> SCALEBITS);
            if(tmp > 255) *output_buf++ = 255;
            else if(tmp < 0) *output_buf++ = 0;
            else *output_buf++ = tmp;
            
            tmp = y + Cbbtab[cb];
            if(tmp > 255) *output_buf++ = 255;
            else if(tmp < 0) *output_buf++ = 0;
            else *output_buf++ = tmp;
        }
        
//        output_buf[outcount++] = (int8) (y);
//        output_buf[outcount++] = (int8) (y);
//        output_buf[outcount++] = (int8) (y);
    }


}



void j3d_init_d_color_deconverter(j3d_decompress_ptr cinfo) {
    my_cconvert_ptr cconvert;

    cconvert = malloc(sizeof(my_color_converter));

    
    /* Allocate and fill in the conversion tables. */
    build_ycc_rgb_table(cconvert);
    build_range_limit_table(cconvert);
    
    /*
     * assign methods
     */
    cconvert->pub.color_deconvert = ycc_rgb_convert;
    
    cinfo->cconvert = (struct j3d_d_color_deconverter *) cconvert;
}


