/* 
 * dmarker.c: marker reader module
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */

#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include "j3dlib.h"


typedef enum {			/* JPEG marker codes */
    M_SOF0  = 0xc0,
    M_SOF1  = 0xc1,
    M_SOF2  = 0xc2,
    M_SOF3  = 0xc3,
  
    M_SOF5  = 0xc5,
    M_SOF6  = 0xc6,
    M_SOF7  = 0xc7,
  
    M_JPG   = 0xc8,
    M_SOF9  = 0xc9,
    M_SOF10 = 0xca,
    M_SOF11 = 0xcb,
  
    M_SOF13 = 0xcd,
    M_SOF14 = 0xce,
    M_SOF15 = 0xcf,
  
    M_DHT   = 0xc4,
  
    M_DAC   = 0xcc,
  
    M_RST0  = 0xd0,
    M_RST1  = 0xd1,
    M_RST2  = 0xd2,
    M_RST3  = 0xd3,
    M_RST4  = 0xd4,
    M_RST5  = 0xd5,
    M_RST6  = 0xd6,
    M_RST7  = 0xd7,
  
    M_SOI   = 0xd8,
    M_EOI   = 0xd9,
    M_SOS   = 0xda,
    M_DQT   = 0xdb,
    M_DNL   = 0xdc,
    M_DRI   = 0xdd,
    M_DHP   = 0xde,
    M_EXP   = 0xdf,
  
    M_APP0  = 0xe0,
    M_APP1  = 0xe1,
    M_APP2  = 0xe2,
    M_APP3  = 0xe3,
    M_APP4  = 0xe4,
    M_APP5  = 0xe5,
    M_APP6  = 0xe6,
    M_APP7  = 0xe7,
    M_APP8  = 0xe8,
    M_APP9  = 0xe9,
    M_APP10 = 0xea,
    M_APP11 = 0xeb,
    M_APP12 = 0xec,
    M_APP13 = 0xed,
    M_APP14 = 0xee,
    M_APP15 = 0xef,
  
    M_JPG0  = 0xf0,
    M_JPG13 = 0xfd,
    M_COM   = 0xfe,
  
    M_TEM   = 0x01,
  
    M_ERROR = 0x100
} J3D_MARKER;


/* Private state */

typedef struct {
    struct j3d_d_marker pub; /* public fields */

//  unsigned int last_restart_interval; /* last DRI value emitted; 0 after SOI */
} my_marker;

typedef my_marker * my_marker_ptr;




/*
 * Macros for fetching data from the data source module.
 *
 * At all times, cinfo->src->next_input_byte and ->bytes_in_buffer reflect
 * the current restart point; we update them only when we have reached a
 * suitable place to restart if a suspension occurs.
 */

/* Declare and initialize local copies of input pointer/count */
#define INPUT_VARS(cinfo)  \
	struct j3d_d_input * datasrc = (cinfo)->input;  \
	int8 * next_input_byte = datasrc->next_input_byte;  \
	unsigned int bytes_in_buffer = datasrc->bytes_in_buffer

/* Unload the local copies --- do this only at a restart boundary */
#define INPUT_SYNC(cinfo)  \
	( datasrc->next_input_byte = next_input_byte,  \
	  datasrc->bytes_in_buffer = bytes_in_buffer )

/* Reload the local copies --- used only in MAKE_BYTE_AVAIL */
#define INPUT_RELOAD(cinfo)  \
	( next_input_byte = datasrc->next_input_byte,  \
	  bytes_in_buffer = datasrc->bytes_in_buffer )

/* Internal macro for INPUT_BYTE and INPUT_2BYTES: make a byte available.
 * Note we do *not* do INPUT_SYNC before calling fill_input_buffer,
 * but we must reload the local copies after a successful fill.
 */
#define MAKE_BYTE_AVAIL(cinfo,action)  \
	if (bytes_in_buffer == 0) {  \
	  if (! (*datasrc->fill_input_buffer) (cinfo))  \
	    { action; }  \
	  INPUT_RELOAD(cinfo);  \
	}

/* Read a byte into variable V.
 * If must suspend, take the specified action (typically "return FALSE").
 */
#define INPUT_BYTE(cinfo,V,action)  \
                  MAKE_BYTE_AVAIL(cinfo,action); \
		  bytes_in_buffer--; \
		  V = *next_input_byte++; 

/* As above, but read two bytes interpreted as an unsigned 16-bit integer.
 * V should be declared unsigned int or perhaps INT32.
 */
#define INPUT_2BYTES(cinfo,V,action)  \
                  MAKE_BYTE_AVAIL(cinfo,action); \
		  bytes_in_buffer--; \
		  V = ((unsigned int) *next_input_byte++) << 8; \
		  MAKE_BYTE_AVAIL(cinfo,action); \
		  bytes_in_buffer--; \
		  V += *next_input_byte++;  

/*
 * Routines to process JPEG markers.
 *
 * Entry condition: JPEG marker itself has been read and its code saved
 *   in cinfo->unread_marker; input restart point is just after the marker.
 *
 * Exit: if return TRUE, have read and processed any parameters, and have
 *   updated the restart point to point after the parameters.
 *   If return FALSE, was forced to suspend before reaching end of
 *   marker parameters; restart point has not been moved.  Same routine
 *   will be called again after application supplies more input data.
 *
 * This approach to suspension assumes that all of a marker's parameters
 * can fit into a single input bufferload.  This should hold for "normal"
 * markers.  Some COM/APPn markers might have large parameter segments
 * that might not fit.  If we are simply dropping such a marker, we use
 * skip_input_data to get past it, and thereby put the problem on the
 * source manager's shoulders.  If we are saving the marker's contents
 * into memory, we use a slightly different convention: when forced to
 * suspend, the marker processor updates the restart point to the end of
 * what it's consumed (ie, the end of the buffer) before returning FALSE.
 * On resumption, cinfo->unread_marker still contains the marker code,
 * but the data source will point to the next chunk of marker data.
 * The marker processor must retain internal state to deal with this.
 *
 * Note that we don't bother to avoid duplicate trace messages if a
 * suspension occurs within marker parameters.  Other side effects
 * require more care.
 */


static int get_soi (j3d_decompress_ptr cinfo)
/* Process an SOI marker */
{
    DEBUGP("SOI\n");
    if (cinfo->marker->saw_SOI) {
        /* multiple SOI are allowed in J3D */
    }
    


    /* Reset all parameters that are defined to be reset by SOI */
    cinfo->restart_interval = 0;

    /* Set initial assumptions for colorspace etc */
/* ToDo */


    
    cinfo->marker->saw_SOI = -1;
    return -1;
}


static int get_sof(j3d_decompress_ptr cinfo)
/* Process a SOFn marker */
{
    int32 length;
    int c, ci, id;
    INPUT_VARS(cinfo);
    
    DEBUGP("SOF: ");

    INPUT_2BYTES(cinfo, length, return 0);

    INPUT_BYTE(cinfo, c, return 0); /* don't care, data precision is fixed for now */
    INPUT_2BYTES(cinfo, cinfo->height, return 0);
    INPUT_2BYTES(cinfo, cinfo->width, return 0);
    INPUT_BYTE(cinfo, cinfo->color_components, return 0);

    DEBUGP("width=%i height=%i color_comp=%i\n",
           cinfo->width, cinfo->height, cinfo->color_components);
    
    length -= 8;

    if (cinfo->marker->saw_SOF) {
        /* not first frame, don't care */
    }


    
    for (ci = 0; ci < cinfo->color_components; ci++) {
        INPUT_BYTE(cinfo, id, return 0);
        if(id >= cinfo->color_components) {
            ERREXIT(-1, "SOF: component-ID must be smaller than number of components");
        }

        /* downsampling factor is the same horizontal as vertical */
        INPUT_BYTE(cinfo, c, return 0);
        if((cinfo->downsampling_factor[id] = (c >> 4) & 15) != ((c     ) & 15)) {
            ERREXIT(-1, "SOF: downsampling factor is not the same horizontal as vertical");
        }
        
        INPUT_BYTE(cinfo, cinfo->quant_tbl_no[id], return 0);
    }

    cinfo->marker->saw_SOF = -1;

    INPUT_SYNC(cinfo);
    return -1;
}


static int get_sos(j3d_decompress_ptr cinfo)
/* Process a SOS marker */
{
    int32 length;
    int i, n, c, cc;
    INPUT_VARS(cinfo);

    DEBUGP("SOS: ");

    if (! cinfo->marker->saw_SOF)
        ERREXIT(-1, "get_sos: sof missing\n");

    INPUT_2BYTES(cinfo, length, return 0);

    INPUT_BYTE(cinfo, n, return 0); /* Number of components */

    if (length != (n * 2 + 6) || n < 1 || n >  MAX_COLOR_COMPONENTS)
        ERREXIT(-1, "get_sos: bad length\n");

    DEBUGP("color_comp=%i ", n);
    cinfo->color_components = n;

    /* Collect the component-spec parameters */

    DEBUGP("id=");
    for (i = 0; i < n; i++) {
        INPUT_BYTE(cinfo, cc, return 0);
        INPUT_BYTE(cinfo, c, return 0);
        DEBUGP("%i, ", cc);
        
        cinfo->dc_huff_tbl_no[cc] = (c >> 4) & 15;
        cinfo->ac_huff_tbl_no[cc] = (c     ) & 15;
    }
    DEBUGP("\b\b  ");
    
    
    /* Collect the additional scan parameters Ss, Se, Ah/Al. */
    INPUT_BYTE(cinfo, c, return 0); /* Ss, not used */
    INPUT_BYTE(cinfo, c, return 0); /* Se, not used */
    INPUT_BYTE(cinfo, c, return 0); /* Ah, Al: not used */

    /* Prepare to scan data & restart markers */
    cinfo->marker->next_restart_num = 0;

    /* Count another SOS marker */
//    cinfo->input_scan_number++;

    DEBUGP("\n");
    
    INPUT_SYNC(cinfo);
    return -1;
}




static int get_dht(j3d_decompress_ptr cinfo)
/* Process a DHT marker */
{
    int32 length;
    unsigned char bits[17];
    unsigned char huffval[256];
    int i, index, count;
    j3d_huff_tbl **htblptr;
    INPUT_VARS(cinfo);
    
    DEBUGP("DHT: ");

    INPUT_2BYTES(cinfo, length, return 0);
    length -= 2;
  
    while (length > 16) {
        INPUT_BYTE(cinfo, index, return 0);


//        DEBUGP("\nbits\n");
        
        bits[0] = 0;
        count = 0;
        for (i = 1; i <= 16; i++) {
            INPUT_BYTE(cinfo, bits[i], return 0);
//            DEBUGP("%02x ", bits[i]);
            count += bits[i];
        }

        length -= 1 + 16;

        /* Here we just do minimal validation of the counts to avoid walking
         * off the end of our table space.  jdhuff.c will check more carefully.
         */
        if (count > 256 || ((int32) count) > length)
            ERREXIT(-1, "get_dht: bad huffman table");

//        DEBUGP("\nval\n");
        for (i = 0; i < count; i++){
            INPUT_BYTE(cinfo, huffval[i], return 0);
//            DEBUGP("%02x ", huffval[i]);
        }
        
        length -= count;

        if (index & 0x10) {		/* AC table definition */
            index -= 0x10;
            DEBUGP("type=AC ");
            
            htblptr = &cinfo->ac_huff_tbl_ptrs[index];
        } else {			/* DC table definition */
            htblptr = &cinfo->dc_huff_tbl_ptrs[index];
            DEBUGP("type=DC ");
        }
        DEBUGP("index=%i ", index);

        if (index < 0 || index >= MAX_HUFF_TBLS)
            ERREXIT(-1, "get_dht: invalid index (%i)\n", index);

        j3d_add_huff_table(htblptr, bits, huffval);
        
/*          if (*htblptr == NULL) */
/*              *htblptr = jpeg_alloc_huff_table((j_common_ptr) cinfo); */
  
/*          MEMCOPY((*htblptr)->bits, bits, SIZEOF((*htblptr)->bits)); */
/*          MEMCOPY((*htblptr)->huffval, huffval, SIZEOF((*htblptr)->huffval)); */
    }

    if (length != 0)
        ERREXIT(-1, "get_dht: bad length\n");

    INPUT_SYNC(cinfo);
    DEBUGP("\n");
    return -1;
}


static int get_dqt(j3d_decompress_ptr cinfo)
/* Process a DQT marker */
{
    int32 length;
    int n, i, prec;
    unsigned int tmp;
    j3d_quant_tbl qtbl;
    const int qtbl_size = DCTSIZE2 * cinfo->max_depth;
    unsigned int * zigzag_order;
    INPUT_VARS(cinfo);    

    DEBUGP("DQT: ");

    
    if(cinfo->max_depth == 8) zigzag_order = cinfo->coef_zigzag_order_8;
    else ERREXIT(-1, "get_dqt: max_depth != 8 not supported yet\n");
                              

    INPUT_2BYTES(cinfo, length, return 0);
    length -= 2;

    while (length > 0) {
        INPUT_BYTE(cinfo, n, return 0);
        prec = n >> 4;
        n &= 0x0F;
        DEBUGP("no=%i", n);

        if (n >= MAX_QUANT_TBLS)
            ERREXIT(-1, "get_dqt: too many tables\n");
      
        if (cinfo->quant_tbl_ptrs[n] == NULL)
            cinfo->quant_tbl_ptrs[n] = (j3d_quant_tbl) calloc(DCT_MAXSIZE3, sizeof(unsigned int));
        
        qtbl = cinfo->quant_tbl_ptrs[n];
        
        for (i = 0; i < qtbl_size; i++) {
            if (prec) {
                INPUT_2BYTES(cinfo, tmp, return 0);
            } else {
                INPUT_BYTE(cinfo, tmp, return 0);
            }
            
            /* We convert the zigzag-order table to natural array order. */
            qtbl[zigzag_order[i]] = (unsigned int) tmp;
        }

#if 0        
        if (cinfo->err->trace_level >= 2) {
            for (i = 0; i < DCTSIZE2; i += 8) {
                TRACEMS8(cinfo, 2, JTRC_QUANTVALS,
                         qtbl->quantval[i],   qtbl->quantval[i+1],
                         qtbl->quantval[i+2], qtbl->quantval[i+3],
                         qtbl->quantval[i+4], qtbl->quantval[i+5],
                         qtbl->quantval[i+6], qtbl->quantval[i+7]);
            }
        }
#endif
        
        length -= qtbl_size+1;
        if (prec) length -= qtbl_size;
    }

    if (length != 0)
        ERREXIT(-1, "get_dqt: bad length\n");

    INPUT_SYNC(cinfo);
    DEBUGP("\n");
    
    return -1;
}


static int get_dri(j3d_decompress_ptr cinfo)
/* Process a DRI marker */
{
    int32 length;
    unsigned int tmp;
    INPUT_VARS(cinfo);

    DEBUGP("DRI: ");

    
    INPUT_2BYTES(cinfo, length, return 0);
  
    if (length != 4)
        ERREXIT(-1, "DRI marker: bad length");

    INPUT_2BYTES(cinfo, tmp, return 0);

    cinfo->restart_interval = tmp;
    DEBUGP("restart_interval=%i\n", cinfo->restart_interval);

    INPUT_SYNC(cinfo);
    return -1;
}






/*
 * Find the next JPEG marker, save it in cinfo->unread_marker.
 * Returns FALSE if had to suspend before reaching a marker;
 * in that case cinfo->unread_marker is unchanged.
 *
 * Note that the result might not be a valid marker code,
 * but it will never be 0 or FF.
 */

static int next_marker (j3d_decompress_ptr cinfo)
{
    int c;
    INPUT_VARS(cinfo);

    for (;;) {
        INPUT_BYTE(cinfo, c, return 0);
        /* Skip any non-FF bytes.
         * This may look a bit inefficient, but it will not occur in a valid file.
         * We sync after each discarded byte so that a suspending data source
         * can discard the byte from its buffer.
         */
        while (c != 0xFF) {
            cinfo->marker->discarded_bytes++;
            INPUT_SYNC(cinfo);
            INPUT_BYTE(cinfo, c, return 0);
        }
        /* This loop swallows any duplicate FF bytes.  Extra FFs are legal as
         * pad bytes, so don't count them in discarded_bytes.  We assume there
         * will not be so many consecutive FF bytes as to overflow a suspending
         * data source's input buffer.
         */
        do {
            INPUT_BYTE(cinfo, c, return 0);
        } while (c == 0xFF);
        if (c != 0)
            break;			/* found a valid marker, exit loop */
        /* Reach here if we found a stuffed-zero data sequence (FF/00).
         * Discard it and loop back to try again.
         */
        cinfo->marker->discarded_bytes += 2;
        INPUT_SYNC(cinfo);
    }

    if (cinfo->marker->discarded_bytes != 0) {
        //WARNMS2(cinfo, JWRN_EXTRANEOUS_DATA, cinfo->marker->discarded_bytes, c);
        DEBUGP("!discarded %i bytes\n", cinfo->marker->discarded_bytes);
        cinfo->marker->discarded_bytes = 0;
    }

    cinfo->unread_marker = c;

    INPUT_SYNC(cinfo);
    return -1;
}


static int first_marker (j3d_decompress_ptr cinfo)
/* Like next_marker, but used to obtain the initial SOI marker. */
/* For this marker, we do not allow preceding garbage or fill; otherwise,
 * we might well scan an entire input file before realizing it ain't JPEG.
 * If an application wants to process non-JFIF files, it must seek to the
 * SOI before calling the JPEG library.
 */
{
    int c, c2;
    INPUT_VARS(cinfo);

    INPUT_BYTE(cinfo, c, return 0);
    INPUT_BYTE(cinfo, c2, return 0);
    if (c != 0xFF || c2 != (int) M_SOI) {
        DEBUGP("!SOI expected, found %02x%02x\n", c, c2);
        ERREXIT(-1, "marker->first_marker: SOI not found. wrong file format.");
        //ERREXIT2(cinfo, JERR_NO_SOI, c, c2);
    }
  


    cinfo->unread_marker = c2;

    INPUT_SYNC(cinfo);
    return -1;
}



/*
 * Read markers until SOS or EOI.
 *
 * Returns same codes as are defined for jpeg_consume_input:
 * JPEG_SUSPENDED, JPEG_REACHED_SOS, or JPEG_REACHED_EOI.
 */

static int read_markers (j3d_decompress_ptr cinfo)
{

//    DEBUGP("processing markers\n");
    
        
    /* Outer loop repeats once for each marker. */
    for (;;) {
        /* Collect the marker proper, unless we already did. */
        /* NB: first_marker() enforces the requirement that SOI appear first. */
        if (cinfo->unread_marker == 0) {
            if (! cinfo->marker->saw_SOI) {
                if (! first_marker(cinfo))
//	  return JPEG_SUSPENDED;
                    return J3D_ERROR;
            } else {
                if (! next_marker(cinfo))
//	  return JPEG_SUSPENDED;
                    return J3D_ERROR;
            }
        }


    
        /* At this point cinfo->unread_marker contains the marker code and the
         * input point is just past the marker proper, but before any parameters.
         * A suspension will cause us to return with this state still true.
         */
        switch (cinfo->unread_marker) {
        case M_SOI:
            if (! get_soi(cinfo))
                return J3D_ERROR;
            break;
            
        case M_SOF0:		/* Baseline */
        case M_SOF1:		/* Extended sequential, Huffman */
            if (! get_sof(cinfo))
                return J3D_ERROR;
            break;

            /* Currently unsupported SOFn types */
        case M_SOF2:		/* Progressive, Huffman */
        case M_SOF9:		/* Extended sequential, arithmetic */
        case M_SOF10:		/* Progressive, arithmetic */
        case M_SOF3:		/* Lossless, Huffman */
        case M_SOF5:		/* Differential sequential, Huffman */
        case M_SOF6:		/* Differential progressive, Huffman */
        case M_SOF7:		/* Differential lossless, Huffman */
        case M_JPG:			/* Reserved for JPEG extensions */
        case M_SOF11:		/* Lossless, arithmetic */
        case M_SOF13:		/* Differential sequential, arithmetic */
        case M_SOF14:		/* Differential progressive, arithmetic */
        case M_SOF15:		/* Differential lossless, arithmetic */
            ERREXIT(-1, "unsupported SOFn type (0x%02x)\n", cinfo->unread_marker);
            break;

        case M_SOS:
            if (! get_sos(cinfo))
                return J3D_ERROR;
            cinfo->unread_marker = 0;	/* processed the marker */
            return J3D_REACHED_SOS;
            
        case M_EOI:
            cinfo->unread_marker = 0;	/* processed the marker */
            DEBUGP("EOI reached.\n");
            return J3D_REACHED_EOI;
#if 0
        case M_DAC:
            if (! get_dac(cinfo))
                return JPEG_SUSPENDED;
            break;
      
#endif       
        case M_DHT:
            if (! get_dht(cinfo))
                return J3D_ERROR;
            break;
        case M_DQT:
            if (! get_dqt(cinfo))
                return J3D_ERROR;
            break;

        case M_DRI:
            if (! get_dri(cinfo))
                return J3D_ERROR;
            break;
#if 0
        case M_APP0:
        case M_APP1:
        case M_APP2:
        case M_APP3:
        case M_APP4:
        case M_APP5:
        case M_APP6:
        case M_APP7:
        case M_APP8:
        case M_APP9:
        case M_APP10:
        case M_APP11:
        case M_APP12:
        case M_APP13:
        case M_APP14:
        case M_APP15:
            if (! (*((my_marker_ptr) cinfo->marker)->process_APPn[
		cinfo->unread_marker - (int) M_APP0]) (cinfo))
                return JPEG_SUSPENDED;
            break;
      
        case M_COM:
            if (! (*((my_marker_ptr) cinfo->marker)->process_COM) (cinfo))
                return JPEG_SUSPENDED;
            break;

#endif
        case M_RST0:		/* these are all parameterless */
        case M_RST1:
        case M_RST2:
        case M_RST3:
        case M_RST4:
        case M_RST5:
        case M_RST6:
        case M_RST7:
        case M_TEM:
            break;
#if 0
        case M_DNL:			/* Ignore DNL ... perhaps the wrong thing */
            if (! skip_variable(cinfo))
                return JPEG_SUSPENDED;
            break;
#endif
        default:			/* must be DHP, EXP, JPGn, or RESn */
            /* For now, we treat the reserved markers as fatal errors since they are
             * likely to be used to signal incompatible JPEG Part 3 extensions.
             * Once the JPEG 3 version-number marker is well defined, this code
             * ought to change!
             */
            ERREXIT(-1, "wrong file format (unknown marker 0x%02x found)\n", cinfo->unread_marker);
            break;
        }
        /* Successfully processed marker, so reset state variable */

        cinfo->unread_marker = 0;
    } /* end loop */
}

/*
 * This is the default resync_to_restart method for data source managers
 * to use if they don't have any better approach.  Some data source managers
 * may be able to back up, or may have additional knowledge about the data
 * which permits a more intelligent recovery strategy; such managers would
 * presumably supply their own resync method.
 *
 * read_restart_marker calls resync_to_restart if it finds a marker other than
 * the restart marker it was expecting.  (This code is *not* used unless
 * a nonzero restart interval has been declared.)  cinfo->unread_marker is
 * the marker code actually found (might be anything, except 0 or FF).
 * The desired restart marker number (0..7) is passed as a parameter.
 * This routine is supposed to apply whatever error recovery strategy seems
 * appropriate in order to position the input stream to the next data segment.
 * Note that cinfo->unread_marker is treated as a marker appearing before
 * the current data-source input point; usually it should be reset to zero
 * before returning.
 * Returns FALSE if suspension is required.
 *
 * This implementation is substantially constrained by wanting to treat the
 * input as a data stream; this means we can't back up.  Therefore, we have
 * only the following actions to work with:
 *   1. Simply discard the marker and let the entropy decoder resume at next
 *      byte of file.
 *   2. Read forward until we find another marker, discarding intervening
 *      data.  (In theory we could look ahead within the current bufferload,
 *      without having to discard data if we don't find the desired marker.
 *      This idea is not implemented here, in part because it makes behavior
 *      dependent on buffer size and chance buffer-boundary positions.)
 *   3. Leave the marker unread (by failing to zero cinfo->unread_marker).
 *      This will cause the entropy decoder to process an empty data segment,
 *      inserting dummy zeroes, and then we will reprocess the marker.
 *
 * #2 is appropriate if we think the desired marker lies ahead, while #3 is
 * appropriate if the found marker is a future restart marker (indicating
 * that we have missed the desired restart marker, probably because it got
 * corrupted).
 * We apply #2 or #3 if the found marker is a restart marker no more than
 * two counts behind or ahead of the expected one.  We also apply #2 if the
 * found marker is not a legal JPEG marker code (it's certainly bogus data).
 * If the found marker is a restart marker more than 2 counts away, we do #1
 * (too much risk that the marker is erroneous; with luck we will be able to
 * resync at some future point).
 * For any valid non-restart JPEG marker, we apply #3.  This keeps us from
 * overrunning the end of a scan.  An implementation limited to single-scan
 * files might find it better to apply #2 for markers other than EOI, since
 * any other marker would have to be bogus data in that case.
 */

static int resync_to_restart(j3d_decompress_ptr cinfo, int desired)
{
    int marker = cinfo->unread_marker;
    int action = 1;
  
    /* Always put up a warning. */
    DEBUGP("resync: marker=%02x -> ", cinfo->unread_marker);
  
    /* Outer loop handles repeated decision after scanning forward. */
    for (;;) {
        if (marker < (int) M_SOF0)
            action = 2;		/* invalid marker */
        else if (marker < (int) M_RST0 || marker > (int) M_RST7)
            action = 3;		/* valid non-restart marker */
        else {
            if (marker == ((int) M_RST0 + ((desired+1) & 7)) ||
                marker == ((int) M_RST0 + ((desired+2) & 7)))
                action = 3;		/* one of the next two expected restarts */
            else if (marker == ((int) M_RST0 + ((desired-1) & 7)) ||
                     marker == ((int) M_RST0 + ((desired-2) & 7)))
                action = 2;		/* a prior restart, so advance */
            else
                action = 1;		/* desired restart or too far away */
        }

        switch (action) {
        case 1:
            /* Discard marker and let entropy decoder resume processing. */
            DEBUGP("discarding marker\n");
            cinfo->unread_marker = 0;
            return -1;
        case 2:
            /* Scan to the next marker, and repeat the decision loop. */
            DEBUGP("scan to next marker\n");
            if (! next_marker(cinfo))
                return 0;
            marker = cinfo->unread_marker;
            break;
        case 3:
            /* Return without advancing past this marker. */
            /* Entropy decoder will be forced to process an empty segment. */
            DEBUGP("process empty segment\n");
            return -1;
        }
    } /* end loop */
}


/*
 * Read a restart marker, which is expected to appear next in the datastream;
 * if the marker is not there, take appropriate recovery action.
 * Returns FALSE if suspension is required.
 *
 * This is called by the entropy decoder after it has read an appropriate
 * number of MCUs.  cinfo->unread_marker may be nonzero if the entropy decoder
 * has already read a marker from the data source.  Under normal conditions
 * cinfo->unread_marker will be reset to 0 before returning; if not reset,
 * it holds a marker which the decoder will be unable to read past.
 */

static int read_restart_marker(j3d_decompress_ptr cinfo)
{
    /* Obtain a marker unless we already did. */
    /* Note that next_marker will complain if it skips any data. */
    if (cinfo->unread_marker == 0) {
        if (! next_marker(cinfo))
            return 0;
    }

    if (cinfo->unread_marker ==
        ((int) M_RST0 + cinfo->marker->next_restart_num)) {
        /* Normal case --- swallow the marker and let entropy decoder continue */
        cinfo->unread_marker = 0;
    } else {
        /* Uh-oh, the restart markers have been messed up. */
        /* Let the data source manager determine how to resync. */
        DEBUGP("!!! messed up markers -> resync !!!\n");
        if (! resync_to_restart(cinfo, cinfo->marker->next_restart_num))        
            return 0;
    }

    /* Update next-restart state */
    cinfo->marker->next_restart_num = (cinfo->marker->next_restart_num + 1) & 7;

    return -1;
}





/*
 * Initialize the marker reader module.
 * This is called only once, when the decompression object is created.
 */

void j3d_init_d_marker (j3d_decompress_ptr cinfo)
{
    my_marker_ptr marker;

    /* Create the subobject */
    marker = (my_marker_ptr) malloc(sizeof(my_marker));


    marker->pub.saw_SOI = 0;
    marker->pub.saw_SOF = 0;
    marker->pub.next_restart_num = 0;
    marker->pub.discarded_bytes = 0;
  
    cinfo->marker = (struct j3d_d_marker *) marker;


    
    /* Initialize method pointers */
    marker->pub.read_markers = read_markers;
    marker->pub.read_restart_marker = read_restart_marker;
    //  marker->pub.write_frame_header = write_frame_header;
    //  marker->pub.write_scan_header = write_scan_header;
//  marker->pub.write_file_trailer = write_file_trailer;
    //  marker->pub.write_tables_only = write_tables_only;
    //  marker->pub.write_marker_header = write_marker_header;
    //  marker->pub.write_marker_byte = write_marker_byte;
    
    /* Initialize private state */
//  marker->last_restart_interval = 0;
}

