/* 
 * dmcu.c: MCU manager (decompress) module
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */

/* 
 * ToDo: there is some code in common with cmcu.c
 */

#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include "j3dlib.h"


/* current state for MCU manager */
typedef struct {
    j3d_dct_coef_ptr coef;                      /* current coef-cubes */
    int mcu_length;                             /* amount of cubes per MCU */
    int total_mcu;                              /* total amount of MCU's  per sequence */
    int curr_mcu;                               /* current MCU */
    int curr_pos;                               /* current position in MCU */
    int component[MCU_MAXLENGTH];               /* color component of current cube */
    int comp_width[MAX_COLOR_COMPONENTS];           /* width (=height) of component in an MCU */
    int dist[MCU_MAXLENGTH];                    /* distance to next cube (of same component in MCU). if 0: last of MCU */
    int mcu_ptr[MAX_COLOR_COMPONENTS];
    int mcu_ptr_off[MAX_COLOR_COMPONENTS];
} mcu_current_state;



/* Private subobject */
typedef struct {
    struct j3d_d_mcu pub; /* public fields */

    /* private variables */

    mcu_current_state state;
} my_mcu;
typedef my_mcu * my_mcu_ptr;


/*
 * return values: J3D_MCU_OK, J3D_MCU_COMPLETE, J3D_MCU_SEQ_COMPLETE, J3D_MCU_NO_MORE_BLOCKS
 * ToDo: maybe return a "cube-info"-struct 
 */
static int get_next_cube(j3d_decompress_ptr cinfo, j3d_dct_coefcube *out_cube, unsigned int *out_comp, unsigned int *out_depth) {
    my_mcu_ptr mcu = (my_mcu_ptr) cinfo->mcu;
    mcu_current_state * state = &(mcu->state);
    int curr_comp = state->component[state->curr_pos];
    int retval = 0;

//    DEBUGP("MCU comp=%i; cube=%i", curr_comp, state->mcu_ptr[curr_comp] + state->mcu_ptr_off[curr_comp]);
    
    
    /* some error handling */    
    if(state->curr_mcu >= state->total_mcu) {
        out_cube = NULL;
        return J3D_MCU_NO_MORE_BLOCKS;
    }
    
    /* set the output values */
    *out_cube = state->coef->color_component[curr_comp][state->mcu_ptr[curr_comp] + state->mcu_ptr_off[curr_comp]];
    *out_comp = curr_comp;
    *out_depth = state->coef->depth;

    
    /* advance ptr */
    if(state->dist[state->curr_pos] == 0) { /* advance mcu_ptr to next mcu */
        if((state->mcu_ptr[curr_comp] += state->comp_width[curr_comp]) % state->coef->width[curr_comp] == 0) {
            /* end of row reached, advance rows */
            state->mcu_ptr[curr_comp] += (state->comp_width[curr_comp] - 1) * state->coef->width[curr_comp];
        }
        state->mcu_ptr_off[curr_comp] = 0;
    } else { /* advance mcu_ptr_off to next cube */
        state->mcu_ptr_off[curr_comp] += state->dist[state->curr_pos];
    }


    /* increase curr_pos and return correct values */
    if(++state->curr_pos == state->mcu_length) {
        state->curr_pos = 0;
        if(++state->curr_mcu == state->total_mcu) retval |= J3D_MCU_SEQ_COMPLETE;
        else retval |= J3D_MCU_COMPLETE;
    }
    
    return (retval | J3D_MCU_OK);
}


/*
 * per-sequence init for MCU-generation.
 */
static void start_seq(j3d_decompress_ptr cinfo, j3d_dct_coef_ptr in_coef) {
    my_mcu_ptr mcu = (my_mcu_ptr) cinfo->mcu;
    mcu_current_state * state = &(mcu->state);
    int i;
    
    DEBUGP("MCU per-sequence init\n");    


    state->coef = in_coef;

    /* setup MCU geometry */
    /* ToDo: set the values correct (according to dwonsampling-factor) */

    state->comp_width[0] = 2;
    state->comp_width[1] = 1;
    state->comp_width[2] = 1;
    
    state->mcu_length = 6;

    
    state->component[0] = 0;
    state->dist[0]      = 1;

    state->component[1] = 0;
    state->dist[1]      = in_coef->width[0] - 2 + 1;

    state->component[2] = 0;
    state->dist[2]      = 1;

    state->component[3] = 0;
    state->dist[3]      = 0;

    state->component[4] = 1;
    state->dist[4]      = 0;

    state->component[5] = 2;
    state->dist[5]      = 0;

    state->curr_pos = 0;

    
    /* compute total_mcu */
    state->total_mcu = in_coef->length[0] / 4; /* ToDo... */

    
    /* zero mcu_ptr */
    for(i = 0; i < cinfo->color_components; i++) {
        state->mcu_ptr[i] = 0;
        state->mcu_ptr_off[i] = 0;
    }

    state->curr_mcu = 0;
}




void j3d_init_d_mcu(j3d_decompress_ptr cinfo) {
    my_mcu_ptr mcu;
    
    mcu = (my_mcu_ptr) malloc(sizeof(my_mcu));
    
        
    /* init the objects used */
    

    /*
     * assign methods
     */
    mcu->pub.get_next_cube = get_next_cube;
    mcu->pub.start_seq = start_seq;

    
    cinfo->mcu = (struct j3d_d_mcu *) mcu;
}

    
