/* 
 * j3dutil.c: utilities for both compress and decompress sides
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This program also contains code from the free IJG JPEG codec
 * software by Thomas G. Lane (www.ijg.org)
 *
 */


#define J3D_INTERNALS
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "j3dlib.h"

/*
 * set standard zigzag-order for an 8x8x8 Cube
 */
unsigned int * j3d_std_coef_zigzag() {
  unsigned int * zigzag;
  unsigned int d; /* distance */
  unsigned int i;
  int x, y, z;   

/*
 * jpeg_natural_order[i] is the natural-order position of the i'th element
 * of zigzag order.
 *
 * When reading corrupted data, the Huffman decoders could attempt
 * to reference an entry beyond the end of this array (if the decoded
 * zero run length reaches past the end of the block).  To prevent
 * wild stores without adding an inner-loop test, we put some extra
 * "63"s after the real entries.  This will cause the extra coefficient
 * to be stored in location 63 of the block, not somewhere random.
 * The worst case would be a run-length of 15, which means we need 16
 * fake entries.
 */

  /* ToDo: recalc value 10000 */
  zigzag = (unsigned int*) calloc(DCTSIZE2 * 8 + 10000, sizeof(unsigned int));


  /* 
   * set the index according to the distance d=x+y+z.
   * all coefficients with distance 0 come first (i.e. 0,0,0) then all with
   * distance 1 (1,0,0; 0,1,0; 0,0,1) etc.
   */
  i = 0;
  for(d = 0; d < 8+8+8; d++) {
    x = (d < 8) ? d : 7;
    while(x >= 0) {
      y = (d - x < 8) ? (d - x) : 7;
      while(y >= 0) {
	if(d - x - y < 8) {
	  z = d - x - y;
	  zigzag[x + 8*y + 64*z] = i++;
	}
	y--;
      }
      x--;
    }
  } 

  for(i = DCTSIZE2 * 8; i < DCTSIZE2 * 8 + 10000; i++) {
      zigzag[i] = 511;
  }
  
#ifdef J3D_LOWDEBUG
  DEBUGP("\n\nStandard Zigzag Table:\n-------------------------\n\n");
  for(z = 0; z < 8; z++) {
      for(y = 0; y < 8; y++) {
          for(x = 0; x < 8; x++) {
              DEBUGP("%i, \t", zigzag[x + 8*y + 64*z]);
          }
          DEBUGP("\n");
      }
      DEBUGP("\n");
  }
#endif
  
  return zigzag;
}



/*
 * Huffman table setup routines
 */
void j3d_add_huff_table (j3d_huff_tbl **htblptr, const unsigned char *bits, const unsigned char *val)
/* Define a Huffman table */
{
  int nsymbols, len;

  if (*htblptr == NULL) *htblptr = (j3d_huff_tbl *) malloc(sizeof(j3d_huff_tbl));


  /* Copy the number-of-symbols-of-each-code-length counts */
  memcpy((*htblptr)->bits, bits, (size_t) sizeof((*htblptr)->bits));

  /* Validate the counts.  We do this here mainly so we can copy the right
   * number of symbols from the val[] array, without risking marching off
   * the end of memory.  jchuff.c will do a more thorough test later.
   */
  nsymbols = 0;
  for (len = 1; len <= 16; len++)
    nsymbols += bits[len];
//  if (nsymbols < 1 || nsymbols > 256)
//    ERREXIT(cinfo, JERR_BAD_HUFF_TABLE);

  memcpy((*htblptr)->huffval, val, (size_t) nsymbols * sizeof(unsigned char));

}

